# Copyright 2021 Patrick C. Tapping
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ["APTDevice", "find_device"]

import logging
import asyncio
from threading import Thread
import atexit
import re

import serial
from serial.tools import list_ports, list_ports_common

from .. import protocol as apt
from ..enums import EndPoint, LEDMode

class APTDevice():
    """
    Initialise and open serial device for the ThorLabs APT controller.

    If the ``serial_port`` parameter is ``None`` (default), then an attempt to detect an APT device
    will be performed.
    The first device found will be initialised.
    If multiple devices are present on the system, then the use of the ``serial_number`` parameter
    will specify a particular device by its serial number.
    This is a `regular expression <https://docs.python.org/3/library/re.html>`_ match, for example
    ``serial_number="83"`` would match devices with serial numbers
    starting with 83, while ``serial_number=".*83$"`` would match devices ending in 83.

    :param serial_port: Serial port device the device is connected to.
    :param serial_number: Regular expression matching the serial number of device to search for.
    :param controller: The destination :class:`EndPoint <thorlabs_apt_device.enums.EndPoint>` for the controller.
    :param bays: Tuple of :class:`EndPoint <thorlabs_apt_device.enums.EndPoint>`\\ (s) for the populated controller bays.
    :param channels: Tuple of indices (1-based) for the controller bay's channels.
    """

    def __init__(self, serial_port=None, serial_number="", controller=EndPoint.RACK, bays=(EndPoint.BAY0,), channels=(1,)):
        
        # If serial_port not specified, search for a device
        if serial_port is None:
            serial_port = find_device(serial_number=serial_number)

        # Accept a serial.tools.list_ports.ListPortInfo object (which we may have just found)
        if isinstance(serial_port, list_ports_common.ListPortInfo):
            serial_port = serial_port.device
        
        if serial_port is None:
            raise RuntimeError(f"No Thorlabs APT devices detected with serial_number matching '{serial_number}'.")

        self._log = logging.getLogger(__name__)
        self._log.info(f"Initialising serial port ({serial_port}).")
        # Open and configure serial port settings for ThorLabs APT controllers
        self._port = serial.Serial(serial_port,
                                baudrate=115200,
                                bytesize=serial.EIGHTBITS,
                                parity=serial.PARITY_NONE,
                                stopbits=serial.STOPBITS_ONE,
                                timeout=0.1,
                                rtscts=False)
        self._port.rts = True
        self._port.reset_input_buffer()
        self._port.reset_output_buffer()
        self._port.rts = False
        self._log.info("Opened serial port OK.")

        # APT protocol unpacker for decoding received messages
        self._unpacker = apt.Unpacker(self._port, on_error="warn")

        # ID numbers for controller, bay device and channel identification
        self.controller = controller
        """ID code for the controller message :class:`EndPoint <thorlabs_apt_device.enums.EndPoint>`."""
        self.bays = bays
        """Tuple of ID codes for the controller's card bay :class:`EndPoint <thorlabs_apt_device.enums.EndPoint>`\\ (s)."""
        self.channels = channels
        """Tuple of indexes for the the channels in card bays."""

        # List of functions to call when error notifications recieved
        self._error_callbacks = set()

        # Create a new event loop for ourselves to queue and send commands
        self._loop = asyncio.new_event_loop()

        # Schedule the first check for incoming data on the serial port
        self._loop.call_soon(self._schedule_reads)
        self.read_interval = 0.01
        """Time to wait between read attempts on the serial port, in seconds."""

        # Schedule sending of the "keep alive" acknowledgement commands
        self._loop.call_soon(self._schedule_keepalives)
        self.keepalive_interval = 0.9
        """Time interval between sending of keepalive commands, in seconds."""

        # Request the controller to start sending regular status updates
        for bay in self.bays:
            self._loop.call_soon(self._write, apt.hw_start_updatemsgs(source=EndPoint.HOST, dest=bay))

        # Create a new thread to run the event loop in
        self._thread = Thread(target=self._run_eventloop)
        # Set as daemon thread so it can be killed automatically at program exit
        self._thread.daemon = True
        self._thread.start()

        # Close the serial port at exit in case close() wasn't called
        atexit.register(self._atexit)


    def _run_eventloop(self):
        """
        Entry point for the event loop thread.
        """
        self._log.debug("Starting event loop.")
        asyncio.set_event_loop(self._loop)
        try:
            self._loop.run_forever()
        finally:
            self._loop.close()
            self._loop = None
        self._log.debug("Event loop stopped.")
        self._close_port()


    def _close_port(self):
        """
        Stop status messages and then close the serial port connection to the controller.
        """
        if self._port is not None:
            self._log.debug("Stopping hardware update messages.")
            try:
                for bay in self.bays:
                    self._write(apt.hw_stop_updatemsgs(source=EndPoint.HOST, dest=bay))
                # Don't know if the disconnect message actually does anything, but might as well send it
                self._log.debug("Sending disconnect notification.")
                self._write(apt.hw_disconnect(source=EndPoint.HOST, dest=self.controller))
            except:
                # Something wrong writing to the port, ignore
                self._log.debug("Unable to send disconnect messages.")
                pass
            self._log.info("Closing serial connection.")
            self._port.close()
            self._port = None


    def _atexit(self):
        """
        Catch exit signal and attempt to close everything gracefully.
        """
        # Request the event loop to stop
        self.close()
        # Wait for event loop thread to finish
        self._thread.join()


    def _write(self, command_bytes):
        """
        Write a command out the the serial port.

        :param command_bytes: Command to send to the device as raw byte array.
        """
        #self._log.debug(f"Writing command bytes: {command_bytes}")
        self._port.write(command_bytes)
        self._port.flush()


    def _schedule_reads(self):
        """
        Check for any incoming messages and process them at regular intervals.
        """
        #self._log.debug(f"Checking for data on serial port.")
        for msg in self._unpacker:
            #self._log.debug(f"Received message: {msg}")
            self._process_message(msg)
        # Schedule next check
        self._loop.call_later(self.read_interval, self._schedule_reads)


    def _schedule_keepalives(self):
        """
        Send the "keep alive" acknowledgement command at regular intervals.
        """
        #self._log.debug(f"Sending keep alive acknowledgement.")
        for bay in self.bays:
            self._loop.call_soon(self._write, apt.mot_ack_dcstatusupdate(source=EndPoint.HOST, dest=bay))
        # Schedule next send
        self._loop.call_later(self.keepalive_interval, self._schedule_keepalives)


    def _process_message(self, m):
        """
        Process a single response message from the controller.

        :param m: The unpacked message from the controller.
        """
        # TODO: Process any messages common to all APT controllers (which ones?)
        if m.msg == "hw_response":
            # Should there be an error code? The documentation is a little unclear
            self._log.warn(f"Received unknown event notification from APT device {m.source}.")
            for callback in self._error_callbacks:
                callback(source=m.source, msgid=0, code=-1, notes="unknown")
        elif m.msg == "hw_rich_response":
            self._log.warn(f"Received event notification code {m.code} from APT device {m.source}: {m.notes}")
            for callback in self._error_callbacks:
                callback(source=m.source, msgid=m.msgid, code=m.code, notes=m.notes)


    def register_error_callback(self, callback_function):
        """
        Register a function to be called in the case of an error being reported by the device.

        The function passed in should have the signature ``callback_function(source, msgid, code, notes)``,
        where ``source`` is the :class:`enums.EndPoint` where the message originated, ``msgid`` is the 
        type of message which triggered the error (or 0 if unknown or a spontaneous error),
        ``code`` is a numerical error code and ``notes`` is a string description.

        :params callback_function: Function to call in case of device error.
        """
        if callable(callback_function):
            self._error_callbacks.add(callback_function)
        else:
            self._log.warn("Attempted to register a non-callable object as a callback function.")
    

    def unregister_error_callback(self, callback_function):
        """
        Unregister a previously registered error callback function.

        The function passed in should have been previously registered using :meth:`register_error_callback`.

        :params callback_function: Function to unregister.
        """
        if callback_function not in self._error_callbacks:
            self._log.warn("Attemped to unregister an unknown function.")
        else:
            self._error_callbacks.pop(callback_function)


    def close(self):
        """
        Close the serial connection to the ThorLabs APT controller.
        """
        if self._loop is not None:
            self._log.debug("Stopping event loop.")
            self._loop.call_soon_threadsafe(self._loop.stop)
        # Note, this returns before event loop has actually stopped and serial port closed


    def identify(self, channel=0):
        """
        Flash the device's front panel LEDs to identify the unit.

        For some single-channel USB controlled devices ``channel=None`` is used,
        which sends the identify command to the USB controller :class:`EndPoint`.
        On devices which are considered "rack" controllers (including single-channel "rack" 
        units such as the BBD201), the ``channel`` parameter will actually refer to the card bay.
        There are likely other types of units (though currently untested) which have a single "bay"
        with multiple channels, and then the ``channel`` parameter would refer to the actual
        channel index of the controller card.

        :param channel: Index (0-based) of controller bay channel to send the command.
        """
        if channel is None:
            self._log.debug("Identifying [channel=None].")
            self._loop.call_soon_threadsafe(self._write, apt.mod_identify(source=EndPoint.HOST, dest=EndPoint.USB, chan_ident=0))
        else:
            self._log.debug(f"Identifying [channel={self.channels[channel]}].")
            self._loop.call_soon_threadsafe(self._write, apt.mod_identify(source=EndPoint.HOST, dest=EndPoint.RACK, chan_ident=self.channels[channel]))
 

def find_device(serial_number=""):
    """
    Search attached serial ports for a Thorlabs APT controller.

    The first device found will be returned.
    If multiple devices are attached to the system, the ``serial_number`` parameter may be used
    to select the correct device. This is a regular expression match, for example
    ``serial_number="83"`` would match devices with serial numbers starting with 83, while
    ``serial_number=".*83$"`` would match devices ending in 83.

    Note that the APT protocol documentation lists formats for device serial numbers.
    For example, a TDC001 "DC Driver T-Cube" should have serial numbers starting with "83".

    :param serial_number: Regular expression to match a device serial number.
    """
    for p in serial.tools.list_ports.comports():
        # If manufacturer and product fields exist, try to use them
        # We require a match on serial number though
        if ((re.match("Thorlabs", p.manufacturer) if p.manufacturer else True)
            and (re.match("APT", p.product) if p.product else True)
            and (re.match(serial_number, p.serial_number) if p.serial_number else False)):
            return p


# Copyright 2021 Patrick C. Tapping
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ["TDC001"]

from .. import protocol as apt
from .aptdevice_dc import APTDevice_DC
from ..enums import EndPoint, LEDMode

class TDC001(APTDevice_DC):
    """
    A class specific to a particular ThorLabs APT device model.

    It is based off :class:`APTDevice_DC` with some customisation for the specifics of the device.
    For example, the controller is single bay/channel, has inverted direction logic, and has a
    few extra device-specific commands.

    Additionally, as it is a single bay/channel controller, aliases of ``status = status_[0][0]``
    etc are created for convenience.

    :param serial_port: Serial port device the device is connected to.
    :param serial_number: Regular expression matching the serial number of device to search for.
    :param home: Perform a homing operation on initialisation.
    :param invert_direction_logic: Invert the meaning of "forward" and "reverse".
    """
    def __init__(self, serial_port=None, serial_number="83", home=True, invert_direction_logic=True, swap_limit_switches=True):
        super().__init__(serial_port=serial_port, serial_number=serial_number, home=home, invert_direction_logic=invert_direction_logic, swap_limit_switches=swap_limit_switches, controller=EndPoint.RACK, bays=(EndPoint.BAY0,), channels=(1,))
        
        self.status = self.status_[0][0]
        """Alias to first bay/channel of :data:`APTDevice_DC.status_`."""
        
        self.velparams = self.velparams_[0][0]
        """Alias to first bay/channel of :data:`APTDevice_DC.velparams_`"""
        
        self.genmoveparams = self.genmoveparams_[0][0]
        """Alias to first bay/channel of :data:`APTDevice_DC.genmoveparams_`"""
        
        self.jogparams = self.jogparams_[0][0]
        """Alias to first bay/channel of :data:`APTDevice_DC.jogparams_`"""
        
        self.pidparams = {
            "proportional" : 0,
            "integral" : 0,
            "differential" : 0,
            "integral_limits" : 0,
            "filter_control" : 0,
            # Update message fields
            "msg" : "",
            "msgid" : 0,
            "source" : 0,
            "dest" : 0,
            "chan_ident" : 0,
        }
        """
        Dictionary of PID algorithm parameters.
        
        Keys are ``"proportional"``, ``"integral"``, ``"differential"``, ``"integral_limits"``, and
        ``"filter_control"``.
        """
        # Request current PID parameters
        self._loop.call_soon_threadsafe(self._write, apt.mot_req_dcpidparams(source=EndPoint.HOST, dest=self.bays[0], chan_ident=self.channels[0]))

        self.ledmode = {
            LEDMode.IDENT : False,
            LEDMode.LIMITSWITCH : False,
            LEDMode.MOVING : False,
        }
        """
        Dictionary describing the LED activity mode.
        
        See :class:`LEDMode <thorlabs_apt_device.enums.LEDMode>` for details on the different modes.
        """
        # Request current LED modes
        self._loop.call_soon_threadsafe(self._write, apt.mot_req_avmodes(source=EndPoint.HOST, dest=self.bays[0], chan_ident=self.channels[0]))

    def _process_message(self, m):
        super()._process_message(m)
        if m.msg == "mot_get_dcpidparams":
            self.pidparams.update(m._asdict())
        elif m.msg == "mot_get_avmodes":
            # LED mode parameter update
            self.ledmode.update({
                LEDMode.IDENT : bool(m.mode_bits & LEDMode.IDENT),
                LEDMode.LIMITSWITCH : bool(m.mode_bits & LEDMode.LIMITSWITCH),
                LEDMode.MOVING : bool(m.mode_bits & LEDMode.MOVING),
            })


    def set_led_mode(self, mode_bits, bay=0, channel=0):
        """
        Configure the behaviour of the controller's status LEDs.

        The ``mode_bits`` can be generated by composing values from the :class:`LEDMode <thorlabs_apt_device.enums.LEDMode>` enum.
        For example, ``LEDMode.IDENT | LEDMode.MOVING`` would set the LED to flash when the identify message is sent, and also when the motor is moving.

        :param mode_bits: Integer containing the relevant mode bits.
        :param bay: Index (0-based) of controller bay to send the command.
        :param channel: Index (0-based) of controller bay channel to send the command.
        """
        self._log.debug(f"Setting LED mode to mode_bits={mode_bits} [bay={self.bays[bay]:#x}, channel={self.channels[channel]}].")
        self._loop.call_soon_threadsafe(self._write, apt.mot_set_avmodes(source=EndPoint.HOST, dest=self.bays[bay], chan_ident=self.channels[channel], mode_bits=mode_bits))
        # Update status with new LED parameters
        self._loop.call_soon_threadsafe(self._write, apt.mot_req_avmodes(source=EndPoint.HOST, dest=self.bays[bay], chan_ident=self.channels[channel]))



# Copyright 2021 Patrick C. Tapping
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Enumerated values for the ThorLabs APT protocol.
"""

__all__ = ["EndPoint", "LEDMode"]

from enum import IntEnum

class EndPoint(IntEnum):
    """
    Numerical codes representing sources and destinations of messages.
    """
    HOST = 0x01
    RACK = 0x11
    BAY0 = 0x21
    BAY1 = 0x22
    BAY2 = 0x23
    BAY3 = 0x24
    BAY4 = 0x25
    BAY5 = 0x26
    BAY6 = 0x27
    BAY7 = 0x28
    BAY8 = 0x29
    BAY9 = 0x2A
    USB = 0x50

    @property
    def description(self):
        """
        Return a string describing the representation of the endpoint value.
        """
        return {
            EndPoint.HOST : "Host controller (control PC)",
            EndPoint.RACK : "Rack controller, motherboard in a card slot system or comms router board",
            EndPoint.BAY0 : "Bay 0 in a card slot system",
            EndPoint.BAY1 : "Bay 1 in a card slot system",
            EndPoint.BAY2 : "Bay 2 in a card slot system",
            EndPoint.BAY3 : "Bay 3 in a card slot system",
            EndPoint.BAY4 : "Bay 4 in a card slot system",
            EndPoint.BAY5 : "Bay 5 in a card slot system",
            EndPoint.BAY6 : "Bay 6 in a card slot system",
            EndPoint.BAY7 : "Bay 7 in a card slot system",
            EndPoint.BAY8 : "Bay 8 in a card slot system",
            EndPoint.BAY9 : "Bay 9 in a card slot system",
            EndPoint.USB  : "Generic USB hardware unit"
        }[self.value]


class LEDMode(IntEnum):
    """
    Numerical values corresponding to bit fields for device LED behaviour modes.

    These values represent bits in the bit field and can be combined with bitwise logic.
    For example, ``LEDMode.IDENT | LEDMode.MOVING`` would set the LED to flash when the identify
    message is sent, and also when the motor is moving.
    """
    IDENT = 1
    LIMITSWITCH = 2
    MOVING = 8

    @property
    def description(self):
        """
        Return a string describing the representation of the endpoint value.
        """
        return {
            LEDMode.IDENT : "LED will flash when the 'Identify' message is sent",
            LEDMode.LIMITSWITCH : "LED will flash when the motor reaches a forward or reverse limit switch",
            LEDMode.MOVING : "LED is lit when the motor is moving"
        }[self.value]



# MIT License

# Copyright (c) 2020 yaq

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import struct
import functools
from typing import Dict, Any

id_to_func = {}
HEADER_SIZE = 6


def parser(msgid):
    def wrapper(func):
        @functools.wraps(func)
        def inner(data: bytes) -> Dict[str, Any]:
            msgid_read, _, dest, source = struct.unpack_from("<HHBB", data)
            dest = dest & ~0x80
            assert msgid == msgid_read
            ret = {"msg": func.__name__, "msgid": msgid, "dest": dest, "source": source}
            ret.update(func(data))
            return ret

        if msgid in id_to_func:
            raise ValueError(f"Duplicate msgid: {hex(msgid)}")
        id_to_func[msgid] = inner
        return inner

    return wrapper


def _parse_dcstatus(data: bytes) -> Dict[str, Any]:
    # I believe the documentation is wrong and velocity is encoded as a "short" and not a "word"
    # (A stage moving in reverse should return a negative velocity, not a very large positive one!)
    chan_ident, position, velocity, _, status_bits = struct.unpack_from(
        "<HlhHL", data, HEADER_SIZE
    )
    ret = {
        "chan_ident": chan_ident,
        "position": position,
        "velocity": velocity,
    }
    ret.update(_parse_status_bits(status_bits))
    return ret


def _parse_status(data: bytes) -> Dict[str, Any]:
    chan_ident, position, enc_count, status_bits = struct.unpack_from(
        "<HllL", data, HEADER_SIZE
    )
    ret = {
        "chan_ident": chan_ident,
        "position": position,
        "enc_count": enc_count,
    }
    ret.update(_parse_status_bits(status_bits))
    return ret


def _parse_status_bits(status_bits: int) -> Dict[str, Any]:
    # Bitfield
    # Tracking and interlock are the same bit?
    return {
        "forward_limit_switch": bool(status_bits & 0x1),
        "reverse_limit_switch": bool(status_bits & 0x2),
        "moving_forward": bool(status_bits & 0x10),
        "moving_reverse": bool(status_bits & 0x20),
        "jogging_forward": bool(status_bits & 0x40),
        "jogging_reverse": bool(status_bits & 0x80),
        "motor_connected": bool(status_bits & 0x100),
        "homing": bool(status_bits & 0x200),
        "homed": bool(status_bits & 0x400),
        "tracking": bool(status_bits & 0x1000),
        "interlock": bool(status_bits & 0x1000),
        "settled": bool(status_bits & 0x2000),
        "motion_error": bool(status_bits & 0x4000),
        "motor_current_limit_reached": bool(status_bits & 0x1000000),
        "channel_enabled": bool(status_bits & 0x80000000),
    }


def _parse_pz_status_bits(status_bits: int) -> Dict[str, Any]:
    # Bitfield
    return {
        "connected": bool(status_bits & 0x1),
        "zeroed": bool(status_bits & 0x10),
        "zeroing": bool(status_bits & 0x20),
        "strain_gauge_connected": bool(status_bits & 0x100),
        "control_mode": bool(status_bits & 0x400),
        "max_75V": bool(status_bits & 0x1000),
        "max_100V": bool(status_bits & 0x2000),
        "max_150V": bool(status_bits & 0x4000),
        "dig_ins": [bool(status_bits & (1 << (21 + i))) for i in range(8)],
        "active": bool(status_bits & 0x20000000),
        "enabled": bool(status_bits & 0x80000000),
    }


def _parse_nt_status_bits(status_bits: int) -> Dict[str, Any]:
    # Bitfield
    return {
        "tracking": bool(status_bits & 0x1),
        "tracking_with_signal": bool(status_bits & 0x2),
        "tracking_channel_a": bool(status_bits & 0x4),
        "tracking_channel_b": bool(status_bits & 0x8),
        "auto_ranging": bool(status_bits & 0x10),
        "under_read": bool(status_bits & 0x20),
        "over_read": bool(status_bits & 0x40),
        "channel_a_connected": bool(status_bits & 0x10000),
        "channel_b_connected": bool(status_bits & 0x20000),
        "channel_a_enabled": bool(status_bits & 0x40000),
        "channel_b_enabled": bool(status_bits & 0x80000),
        "channel_a_ctrl_mode": bool(status_bits & 0x100000),
        "channel_b_ctrl_mode": bool(status_bits & 0x200000),
    }


def _parse_la_status_bits(status_bits: int) -> Dict[str, Any]:
    # Bitfield
    # Note, the adc on the kls101 is ignored for now
    # KFS 2021-02-19
    units = "mA" if bool(status_bits & 0x10) else ""
    units = "mW" if bool(status_bits & 0x20) else ""
    units = "dBm" if bool(status_bits & 0x40) else ""
    return {
        "output_enabled": bool(status_bits & 0x1),
        "keyswitch_enabled": bool(status_bits & 0x2),
        "control_mode": bool(status_bits & 0x4),
        "safety_interlock": bool(status_bits & 0x8),
        "units": units,
        "dig_ins": [bool(status_bits & 1 << i) for i in range(20, 22)],
        "error": bool(status_bits & 1 << 30),
    }


def _parse_ld_status_bits(status_bits: int) -> Dict[str, Any]:
    # Bitfield
    tia_range = "9uA" if bool(status_bits & 0x10) else ""
    tia_range = "100uA" if bool(status_bits & 0x20) else ""
    tia_range = "0.9mA" if bool(status_bits & 0x40) else ""
    tia_range = "10mA" if bool(status_bits & 0x80) else ""
    return {
        "output_enabled": bool(status_bits & 0x1),
        "keyswitch_enabled": bool(status_bits & 0x2),
        "control_mode": bool(status_bits & 0x4),
        "safety_interlock": bool(status_bits & 0x8),
        "tia_range": tia_range,
        "diode_polarity": bool(status_bits & 0x100),
        "external_sma_enabled": bool(status_bits & 0x200),
        "open_circuit": bool(status_bits & 0x800),
        "psu_voltages_ok": bool(status_bits & 0x1000),
        "tia_range_overlimit": bool(status_bits & 0x2000),
        "tia_range_underlimit": bool(status_bits & 0x4000),
        "signal_generator": bool(status_bits & 0x80000),
        "dig_ins": [bool(status_bits & 1 << i) for i in range(20, 22)],
        "error": bool(status_bits & 1 << 30),
        "high_stability": bool(status_bits & 1 << 31),
    }


def _parse_quad_status_bits(status_bits: int) -> Dict[str, Any]:
    # Bitfield
    return {
        "position_monitoring": bool(status_bits & 0x1),
        "open_loop": bool(status_bits & 0x2),
        "closed_loop": bool(status_bits & 0x4),
    }


def _parse_tec_status_bits(status_bits: int) -> Dict[str, Any]:
    # Bitfield
    display_mode = "temp_actual" if bool(status_bits & 0x10) else ""
    display_mode = "temp_set" if bool(status_bits & 0x20) else ""
    display_mode = "temp_delta" if bool(status_bits & 0x40) else ""
    display_mode = "current" if bool(status_bits & 0x80) else ""
    return {
        "output_enabled": bool(status_bits & 0x1),
        "display_mode": display_mode,
        "error": bool(status_bits & (0x1 << 30)),
    }


def _parse_pzmot_status_bits(status_bits: int) -> Dict[str, Any]:
    # Bitfield
    # Tracking and interlock are the same bit?
    return {
        "forward_limit_switch": bool(status_bits & 0x1),
        "reverse_limit_switch": bool(status_bits & 0x2),
        "moving_forward": bool(status_bits & 0x10),
        "moving_reverse": bool(status_bits & 0x20),
        "jogging_forward": bool(status_bits & 0x40),
        "jogging_reverse": bool(status_bits & 0x80),
        "motor_connected": bool(status_bits & 0x100),
        "homing": bool(status_bits & 0x200),
        "homed": bool(status_bits & 0x400),
        "dig_in1": bool(status_bits & 0x100000),
        "power_ok": bool(status_bits & 0x10000000),
        "active": bool(status_bits & 0x20000000),
        "error": bool(status_bits & 0x40000000),
        "channel_enabled": bool(status_bits & 0x80000000),
    }


@parser(0x0212)
def mod_get_chanenablestate(data: bytes) -> Dict[str, Any]:
    return {"chan_ident": data[2], "enabled": data[3] == 0x01}


@parser(0x0002)
def hw_disconnect(data: bytes) -> Dict[str, Any]:
    return {}


@parser(0x0080)
def hw_response(data: bytes) -> Dict[str, Any]:
    return {}


@parser(0x0081)
def hw_richresponse(data: bytes) -> Dict[str, Any]:
    msg_ident, code, notes = struct.unpack_from("<HH64c", data, HEADER_SIZE)
    return {"msg_ident": msg_ident, "code": code, "notes": notes}


@parser(0x0006)
def hw_get_info(data: bytes) -> Dict[str, Any]:
    (
        serial_number,
        model_number,
        type_,
        *firmware_version,
        _,
        _,
        hw_version,
        mod_state,
        nchs,
    ) = struct.unpack_from("<l8sH4B60sHHH", data, HEADER_SIZE)
    return {
        "serial_number": serial_number,
        "model_number": model_number,
        "type": type_,
        "firmware_version": firmware_version[::-1],
        "hw_version": hw_version,
        "mod_state": mod_state,
        "nchs": nchs,
    }


@parser(0x0061)
def rack_get_bayused(data: bytes) -> Dict[str, Any]:
    return {"bay_ident": data[2], "occupied": data[3] == 0x01}


@parser(0x0066)
def hub_get_bayused(data: bytes) -> Dict[str, Any]:
    bay_ident = data[2]
    if bay_ident == 0xFF:
        bay_ident = -1
    return {"bay_ident": bay_ident}


@parser(0x0226)
def rack_get_statusbits(data: bytes) -> Dict[str, Any]:
    # Bitfield
    (status_bits,) = struct.unpack_from("<L", data, HEADER_SIZE)
    return {
        "dig_outs": [
            bool(status_bits & 0x1),
            bool(status_bits & 0x2),
            bool(status_bits & 0x3),
            bool(status_bits & 0x4),
        ]
    }


@parser(0x0230)
def rack_get_digoutputs(data: bytes) -> Dict[str, Any]:
    # Bitfield
    return {
        "dig_outs": [
            bool(data[2] & 0x1),
            bool(data[2] & 0x2),
            bool(data[2] & 0x3),
            bool(data[2] & 0x4),
        ]
    }


@parser(0x0215)
def mod_get_digoutputs(data: bytes) -> Dict[str, Any]:
    # This differs from 0x0225 and 0x0230, as the number of outputs is not known
    # Bitfield
    return {"bits": data[2]}


@parser(0x0252)
def hw_get_kcubemmilock(data: bytes) -> Dict[str, Any]:
    return {"locked": data[3] == 0x01}


@parser(0x0412)
def mot_get_poscounter(data: bytes) -> Dict[str, Any]:
    chan_ident, position = struct.unpack_from("<Hl", data, HEADER_SIZE)
    return {"chan_ident": chan_ident, "position": position}


@parser(0x040B)
def mot_get_enccounter(data: bytes) -> Dict[str, Any]:
    chan_ident, encoder_count = struct.unpack_from("<Hl", data, HEADER_SIZE)
    return {"chan_ident": chan_ident, "encoder_count": encoder_count}


@parser(0x0415)
def mot_get_velparams(data: bytes) -> Dict[str, Any]:
    chan_ident, min_velocity, acceleration, max_velocity = struct.unpack_from(
        "<H3l", data, HEADER_SIZE
    )
    return {
        "chan_ident": chan_ident,
        "min_velocity": min_velocity,
        "acceleration": acceleration,
        "max_velocity": max_velocity,
    }


@parser(0x0418)
def mot_get_jogparams(data: bytes) -> Dict[str, Any]:
    (
        chan_ident,
        jog_mode,
        step_size,
        min_velocity,
        acceleration,
        max_velocity,
        stop_mode,
    ) = struct.unpack_from("<HH4lH", data, HEADER_SIZE)
    return {
        "chan_ident": chan_ident,
        "jog_mode": jog_mode,
        "step_size": step_size,
        "min_velocity": min_velocity,
        "acceleration": acceleration,
        "max_velocity": max_velocity,
        "stop_mode": stop_mode,
    }


@parser(0x042C)
def mot_get_adcinputs(data: bytes) -> Dict[str, Any]:
    adc_input1, adc_input2 = struct.unpack_from("<HH", data, HEADER_SIZE)
    return {
        "adc_input1": adc_input1 * 5 / 2 ** 15,
        "adc_input2": adc_input2 * 5 / 2 ** 15,
    }


@parser(0x0428)
def mot_get_powerparams(data: bytes) -> Dict[str, Any]:
    chan_ident, rest_factor, move_factor = struct.unpack_from("<3H", data, HEADER_SIZE)
    return {
        "chan_ident": chan_ident,
        "rest_factor": rest_factor,
        "move_factor": move_factor,
    }


@parser(0x043C)
def mot_get_genmoveparams(data: bytes) -> Dict[str, Any]:
    chan_ident, backlash_distance = struct.unpack_from("<Hl", data, HEADER_SIZE)
    return {"chan_ident": chan_ident, "backlash_distance": backlash_distance}


@parser(0x0447)
def mot_get_moverelparams(data: bytes) -> Dict[str, Any]:
    chan_ident, relative_distance = struct.unpack_from("<Hl", data, HEADER_SIZE)
    return {"chan_ident": chan_ident, "relative_distance": relative_distance}


@parser(0x0452)
def mot_get_moveabsparams(data: bytes) -> Dict[str, Any]:
    chan_ident, absolute_position = struct.unpack_from("<Hl", data, HEADER_SIZE)
    return {"chan_ident": chan_ident, "absolute_position": absolute_position}


@parser(0x0442)
def mot_get_homeparams(data: bytes) -> Dict[str, Any]:
    (
        chan_ident,
        home_dir,
        limit_switch,
        home_velocity,
        offset_distance,
    ) = struct.unpack_from("<3Hll", data, HEADER_SIZE)
    return {
        "chan_ident": chan_ident,
        "home_dir": home_dir,
        "limit_switch": limit_switch,
        "home_velocity": home_velocity,
        "offset_distance": offset_distance,
    }


@parser(0x0425)
def mot_get_limswitchparams(data: bytes) -> Dict[str, Any]:
    (
        chan_ident,
        cw_hardlimit,
        ccw_hardlimit,
        cw_softlimit,
        ccw_softlimit,
        soft_limit_mode,
    ) = struct.unpack_from("<3HLLH", data, HEADER_SIZE)
    return {
        "chan_ident": chan_ident,
        "cw_hardlimit": cw_hardlimit,
        "ccw_hardlimit": ccw_hardlimit,
        "cw_softlimit": cw_softlimit,
        "ccw_softlimit": ccw_softlimit,
        "soft_limit_mode": soft_limit_mode,
    }


@parser(0x0444)
def mot_move_homed(data: bytes) -> Dict[str, Any]:
    return {"chan_ident": data[2]}


@parser(0x0464)
def mot_move_completed(data: bytes) -> Dict[str, Any]:
    return _parse_dcstatus(data)


@parser(0x0466)
def mot_move_stopped(data: bytes) -> Dict[str, Any]:
    return _parse_dcstatus(data)


@parser(0x04F6)
def mot_get_bowindex(data: bytes) -> Dict[str, Any]:
    chan_ident, bow_index = struct.unpack_from("<HH", data, HEADER_SIZE)
    return {"chan_ident": chan_ident, "bow_index": bow_index}


@parser(0x04A2)
def mot_get_dcpidparams(data: bytes) -> Dict[str, Any]:
    (
        chan_ident,
        proportional,
        integral,
        differential,
        integral_limits,
        filter_control,
    ) = struct.unpack_from("<H4LH", data, HEADER_SIZE)
    return {
        "chan_ident": chan_ident,
        "proportional": proportional,
        "integral": integral,
        "differential": differential,
        "integral_limits": integral_limits,
        "filter_control": filter_control
    }


@parser(0x04B5)
def mot_get_avmodes(data: bytes) -> Dict[str, Any]:
    chan_ident, mode_bits = struct.unpack_from("<HH", data, HEADER_SIZE)
    # Bitfield
    return {"chan_ident": chan_ident, "mode_bits": mode_bits}


@parser(0x04B2)
def mot_get_potparams(data: bytes) -> Dict[str, Any]:
    chan_ident, zero_wnd, vel1, wnd1, vel2, wnd2, vel3, wnd3, vel4 = struct.unpack_from(
        "<HHlHlHlHl", data, HEADER_SIZE
    )
    return {
        "chan_ident": chan_ident,
        "zero_wnd": zero_wnd,
        "vel1": vel1,
        "wnd1": wnd1,
        "vel2": vel2,
        "wnd2": wnd2,
        "vel3": vel3,
        "wnd3": wnd3,
        "vel4": vel4,
    }


@parser(0x04B8)
def mot_get_buttonparams(data: bytes) -> Dict[str, Any]:
    chan_ident, mode, position1, position2, time_out1, time_out2 = struct.unpack_from(
        "<HHllHH", data, HEADER_SIZE
    )
    return {
        "chan_ident": chan_ident,
        "mode": mode,
        "position1": position1,
        "position2": position2,
        "time_out1": time_out1,
        "time_out2": time_out2,
    }


@parser(0x0491)
def mot_get_dcstatusupdate(data: bytes) -> Dict[str, Any]:
    return _parse_dcstatus(data)


@parser(0x04D9)
def mot_get_positionloopparams(data: bytes) -> Dict[str, Any]:
    (
        chan_ident,
        kp_pos,
        integral,
        i_lim_pos,
        differential,
        kd_time_pos,
        kout_pos,
        kaff_pos,
        pos_err_lim,
        _,
        _,
    ) = struct.unpack_from("<3HL5HL2H", data, HEADER_SIZE)
    return {
        "chan_ident": chan_ident,
        "kp_pos": kp_pos,
        "integral": integral,
        "i_lim_pos": i_lim_pos,
        "differential": differential,
        "kd_time_pos": kd_time_pos,
        "kout_pos": kout_pos,
        "kaff_pos": kaff_pos,
        "pos_err_lim": pos_err_lim,
    }


@parser(0x04DC)
def mot_get_motoroutputparams(data: bytes) -> Dict[str, Any]:
    (
        chan_ident,
        cont_current_lim,
        energy_lim,
        motor_lim,
        motor_bias,
        _,
        _,
    ) = struct.unpack_from("<7H", data, HEADER_SIZE)
    return {
        "chan_ident": chan_ident,
        "cont_current_lim": cont_current_lim,
        "energy_lim": energy_lim,
        "motor_lim": motor_lim,
        "motor_bias": motor_bias,
    }


@parser(0x04E2)
def mot_get_tracksettleparams(data: bytes) -> Dict[str, Any]:
    chan_ident, time, settle_window, track_window, _, _ = struct.unpack_from(
        "<6H", data, HEADER_SIZE
    )
    return {
        "chan_ident": chan_ident,
        "time": time,
        "settle_window": settle_window,
        "track_window": track_window,
    }


@parser(0x04E5)
def mot_get_profilemodeparams(data: bytes) -> Dict[str, Any]:
    chan_ident, mode, jerk, _, _ = struct.unpack_from("<HHLHH", data, HEADER_SIZE)
    return {
        "chan_ident": chan_ident,
        "mode": mode,
        "jerk": jerk,
    }


@parser(0x04E8)
def mot_get_joystickparams(data: bytes) -> Dict[str, Any]:
    (
        chan_ident,
        gear_high_max_vel,
        gear_low_accn,
        gear_high_accn,
        dir_sense,
    ) = struct.unpack_from("<H4LH", data, HEADER_SIZE)
    return {
        "chan_ident": chan_ident,
        "gear_high_max_vel": gear_high_max_vel,
        "gear_low_accn": gear_low_accn,
        "gear_high_accn": gear_high_accn,
        "dir_sense": dir_sense,
    }


@parser(0x04D6)
def mot_get_currentloopoarams(data: bytes) -> Dict[str, Any]:
    (
        chan_ident,
        phase,
        kp_current,
        ki_current,
        i_lim_current,
        i_dead_band,
        kff,
        _,
        _,
    ) = struct.unpack_from("<9H", data, HEADER_SIZE)
    return {
        "chan_ident": chan_ident,
        "phase": phase,
        "kp_current": kp_current,
        "ki_current": ki_current,
        "i_lim_current": i_lim_current,
        "i_dead_band": i_dead_band,
        "kff": kff,
    }


@parser(0x04EB)
def mot_get_settledcurrentloopparams(data: bytes) -> Dict[str, Any]:
    (
        chan_ident,
        phase,
        kp_settled,
        ki_settled,
        i_lim_settled,
        i_dead_band_settled,
        kff_settled,
        _,
        _,
    ) = struct.unpack_from("<9H", data, HEADER_SIZE)
    return {
        "chan_ident": chan_ident,
        "phase": phase,
        "kp_settled": kp_settled,
        "ki_settled": ki_settled,
        "i_lim_settled": i_lim_settled,
        "i_dead_band_settled": i_dead_band_settled,
        "kff_settled": kff_settled,
    }


@parser(0x04F2)
def mot_get_stageaxisparams(data: bytes) -> Dict[str, Any]:
    (
        chan_ident,
        stage_id,
        axis_id,
        part_no_axis,
        serial_num,
        counts_per_unit,
        min_pos,
        max_pos,
        max_accn,
        max_dec,
        max_vel,
        _,
        _,
        _,
        _,
        _,
        _,
        _,
        _,
    ) = struct.unpack_from("<HHH16sLL5l4H4L", data, HEADER_SIZE)
    return {
        "chan_ident": chan_ident,
        "stage_id": stage_id,
        "axis_id": axis_id,
        "part_no_axis": part_no_axis,
        "serial_num": serial_num,
        "counts_per_unit": counts_per_unit,
        "min_pos": min_pos,
        "max_pos": max_pos,
        "max_accn": max_accn,
        "max_dec": max_dec,
        "max_vel": max_vel,
    }


@parser(0x0481)
def mot_get_statusupdate(data: bytes) -> Dict[str, Any]:
    return _parse_status(data)


@parser(0x042A)
def mot_get_statusbits(data: bytes) -> Dict[str, Any]:
    chan_ident, status_bits = struct.unpack_from("<HL", data, HEADER_SIZE)
    ret = {
        "chan_ident": chan_ident,
    }
    ret.update(_parse_status_bits(status_bits))
    # Bitfield
    return ret


@parser(0x0502)
def mot_get_trigger(data: bytes) -> Dict[str, Any]:
    return {
        "chan_ident": data[2],
        "mode": data[3],
    }


@parser(0x0522)
def mot_get_kcubemmiparams(data: bytes) -> Dict[str, Any]:
    (
        chan_ident,
        mode,
        max_vel,
        accn,
        dir_sense,
        pre_set_pos1,
        pre_set_pos2,
        disp_brightness,
        disp_timeout,
        disp_dim_level,
    ) = struct.unpack_from("<HHllHll3H", data, HEADER_SIZE)
    return {
        "chan_ident": chan_ident,
        "mode": mode,
        "max_vel": max_vel,
        "accn": accn,
        "dir_sense": dir_sense,
        "pre_set_pos1": pre_set_pos1,
        "pre_set_pos2": pre_set_pos2,
        "disp_brightness": disp_brightness,
        "disp_timeout": disp_timeout,
        "disp_dim_level": disp_dim_level,
    }


@parser(0x0525)
def mot_get_kcubetrigconfig(data: bytes) -> Dict[str, Any]:
    (
        chan_ident,
        trig1_mode,
        trig1_polarity,
        trig2_mode,
        trig2_polarity,
    ) = struct.unpack_from("<5H", data, HEADER_SIZE)
    return {
        "chan_ident": chan_ident,
        "trig1_mode": trig1_mode,
        "trig1_polarity": trig1_polarity,
        "trig2_mode": trig2_mode,
        "trig2_polarity": trig2_polarity,
    }


@parser(0x0528)
def mot_get_kcubeposttrigparams(data: bytes) -> Dict[str, Any]:
    (
        chan_ident,
        start_pos_fwd,
        interval_fwd,
        num_pulses_fwd,
        start_pos_rev,
        interval_rev,
        num_pulses_rev,
        pulse_width,
        num_cycles,
    ) = struct.unpack_from("<H8l", data, HEADER_SIZE)
    return {
        "chan_ident": chan_ident,
        "start_pos_fwd": start_pos_fwd,
        "interval_fwd": interval_fwd,
        "num_pulses_fwd": num_pulses_fwd,
        "start_pos_rev": start_pos_rev,
        "interval_rev": interval_rev,
        "num_pulses_rev": num_pulses_rev,
        "pulse_width": pulse_width,
        "num_cycles": num_cycles,
    }


@parser(0x052B)
def mot_get_kcubekstloopparams(data: bytes) -> Dict[str, Any]:
    (
        chan_ident,
        loop_mode,
        prop,
        int,
        diff,
        pid_clip,
        pid_tol,
        encoder_const,
    ) = struct.unpack_from("<HH5lL", data, HEADER_SIZE)
    return {
        "chan_ident": chan_ident,
        "loop_mode": loop_mode,
        "prop": prop,
        "int": int,
        "diff": diff,
        "pid_clip": pid_clip,
        "pid_tol": pid_tol,
        "encoder_const": encoder_const,
    }


@parser(0x0512)
def mot_get_mmf_operparams(data: bytes) -> Dict[str, Any]:
    (
        chan_ident,
        i_tranit_time,
        i_transit_time_adc,
        oper_mode1,
        sig_mode1,
        pulse_width1,
        oper_mode2,
        sig_mode2,
        pulse_width2,
        _,
        _,
    ) = struct.unpack_from("<HllHHlHHllL", data, HEADER_SIZE)
    return {
        "chan_ident": chan_ident,
        "i_tranit_time": i_tranit_time,
        "i_transit_time_adc": i_transit_time_adc,
        "oper_mode1": oper_mode1,
        "sig_mode1": sig_mode1,
        "pulse_width1": pulse_width1,
        "oper_mode2": oper_mode2,
        "sig_mode2": sig_mode2,
        "pulse_width2": pulse_width2,
    }


@parser(0x04C2)
def mot_get_sol_operatingmode(data: bytes) -> Dict[str, Any]:
    return {"chan_ident": data[2], "mode": data[3]}


@parser(0x04C5)
def mot_get_sol_cycleparams(data: bytes) -> Dict[str, Any]:
    chan_ident, off_time, num_cycles = struct.unpack_from("<H3l", data, HEADER_SIZE)
    return {
        "chan_ident": chan_ident,
        "off_time": off_time,
        "num_cycles": num_cycles,
    }


@parser(0x04C8)
def mot_get_sol_interlockmode(data: bytes) -> Dict[str, Any]:
    return {"chan_ident": data[2], "mode": data[3] == 0x01}


@parser(0x04CD)
def mot_get_sol_state(data: bytes) -> Dict[str, Any]:
    return {"chan_ident": data[2], "state": data[3] == 0x01}


@parser(0x0642)
def pz_get_positioncontrolmode(data: bytes) -> Dict[str, Any]:
    return {"chan_ident": data[2], "mode": data[3]}


@parser(0x0645)
def pz_get_outputvolts(data: bytes) -> Dict[str, Any]:
    chan_ident, voltage = struct.unpack_from("<Hh", data, HEADER_SIZE)
    return {
        "chan_ident": chan_ident,
        "voltage": voltage,
    }


@parser(0x0648)
def pz_get_outputpos(data: bytes) -> Dict[str, Any]:
    chan_ident, position = struct.unpack_from("<HH", data, HEADER_SIZE)
    return {
        "chan_ident": chan_ident,
        "position": position,
    }


@parser(0x0654)
def pz_get_inputvoltssrc(data: bytes) -> Dict[str, Any]:
    chan_ident, volt_src = struct.unpack_from("<HH", data, HEADER_SIZE)
    return {
        "chan_ident": chan_ident,
        "volt_src": volt_src,
    }


@parser(0x0657)
def pz_get_piconsts(data: bytes) -> Dict[str, Any]:
    chan_ident, PropConst, IntConst = struct.unpack_from("<HHH", data, HEADER_SIZE)
    return {
        "chan_ident": chan_ident,
        "PropConst": PropConst,
        "IntConst": IntConst,
    }


@parser(0x065C)
def pz_get_pzstatusbits(data: bytes) -> Dict[str, Any]:
    chan_ident, status_bits = struct.unpack_from("<HL", data, HEADER_SIZE)
    ret = {"chan_ident": chan_ident}
    ret.update(_parse_pz_status_bits(status_bits))
    return ret


@parser(0x0661)
def pz_get_pzstatusupdate(data: bytes) -> Dict[str, Any]:
    chan_ident, output_voltage, position, status_bits = struct.unpack_from(
        "<HhhL", data, HEADER_SIZE
    )
    ret = {
        "chan_ident": chan_ident,
        "output_voltage": output_voltage,
        "position": position,
    }
    ret.update(_parse_pz_status_bits(status_bits))
    return ret


@parser(0x0692)
def pz_get_ppc_pidconsts(data: bytes) -> Dict[str, Any]:
    (
        chan_ident,
        pid_p,
        pid_i,
        pid_d,
        pid_d_filter_cut,
        pid_d_filter_on,
    ) = struct.unpack_from("<HHHHHH", data, HEADER_SIZE)
    return {
        "chan_ident": chan_ident,
        "pid_p": pid_p,
        "pid_i": pid_i,
        "pid_d": pid_d,
        "pid_d_filter_cut": pid_d_filter_cut,
        "pid_d_filter_on": pid_d_filter_on,
    }


@parser(0x0695)
def pz_get_ppc_notchparams(data: bytes) -> Dict[str, Any]:
    (
        chan_ident,
        filter_no,
        filter1_center,
        filter1_q,
        notch_filter_1_on,
        filter2_center,
        filter2_q,
        notch_filter_2_on,
    ) = struct.unpack_from("<HHHHHHHH", data, HEADER_SIZE)
    return {
        "chan_ident": chan_ident,
        "filter_no": filter_no,
        "filter1_center": filter1_center,
        "filter1_q": filter1_q,
        "notch_filter_1_on": notch_filter_1_on,
        "filter2_center": filter2_center,
        "filter2_q": filter2_q,
        "notch_filter_2_on": notch_filter_2_on,
    }


@parser(0x0698)
def pz_get_ppc_iosettings(data: bytes) -> Dict[str, Any]:
    (
        chan_ident,
        control_src,
        monitor_out_sig,
        monitor_out_bandwidth,
        feedback_src,
        fp_brightness,
        _,
    ) = struct.unpack_from("<HHHHHHH", data, HEADER_SIZE)
    return {
        "chan_ident": chan_ident,
        "control_src": control_src,
        "monitor_out_sig": monitor_out_sig,
        "monitor_out_bandwidth": monitor_out_bandwidth,
        "feedback_src": feedback_src,
        "fp_brightness": fp_brightness,
    }


@parser(0x0702)
def pz_get_outputlut(data: bytes) -> Dict[str, Any]:
    chan_ident, index, output = struct.unpack_from("<HHh", data, HEADER_SIZE)
    return {
        "chan_ident": chan_ident,
        "index": index,
        "output": output,
    }


@parser(0x0705)
def pz_get_outputlutparams(data: bytes) -> Dict[str, Any]:
    (
        chan_ident,
        mode,
        cycle_length,
        num_cycles,
        delay_time,
        pre_cycle_rest,
        post_cycle_rest,
        output_trig_start,
        output_trig_width,
        trig_rep_cycle,
    ) = struct.unpack_from("<HHHLLLLHLH", data, HEADER_SIZE)
    return {
        "chan_ident": chan_ident,
        "mode": mode,
        "cycle_length": cycle_length,
        "num_cycles": num_cycles,
        "delay_time": delay_time,
        "pre_cycle_rest": pre_cycle_rest,
        "post_cycle_rest": post_cycle_rest,
        "output_trig_start": output_trig_start,
        "output_trig_width": output_trig_width,
        "trig_rep_cycle": trig_rep_cycle,
    }


@parser(0x07D3)
def pz_get_tpz_dispsettings(data: bytes) -> Dict[str, Any]:
    disp_intensity = struct.unpack_from("<H", data, HEADER_SIZE)
    return {
        "disp_intensity": disp_intensity,
    }


@parser(0x07D6)
def pz_get_tpz_iosettings(data: bytes) -> Dict[str, Any]:
    chan_ident, voltage_limit, hub_analog_input, _, _ = struct.unpack_from(
        "<HHHHH", data, HEADER_SIZE
    )
    return {
        "chan_ident": chan_ident,
        "voltage_limit": voltage_limit,
        "hub_analog_input": hub_analog_input,
    }


@parser(0x0651)
def pz_get_maxtravel(data: bytes) -> Dict[str, Any]:
    chan_ident, travel = struct.unpack_from("<HH", data, HEADER_SIZE)
    return {
        "chan_ident": chan_ident,
        "travel": travel,
    }


@parser(0x0672)
def pz_get_iosettings(data: bytes) -> Dict[str, Any]:
    (
        chan_ident,
        amp_current_limit,
        amp_lowpass_filter,
        feedback_sig,
        bnc_trig_or_lv_output,
    ) = struct.unpack_from("<HHHHH", data, HEADER_SIZE)
    return {
        "chan_ident": chan_ident,
        "amp_current_limit": amp_current_limit,
        "amp_lowpass_filter": amp_lowpass_filter,
        "feedback_sig": feedback_sig,
        "bnc_trig_or_lv_output": bnc_trig_or_lv_output,
    }


@parser(0x0682)
def pz_get_outputmaxvolts(data: bytes) -> Dict[str, Any]:
    chan_ident, voltage, flags = struct.unpack_from("<HHH", data, HEADER_SIZE)
    ret = {
        "chan_ident": chan_ident,
        "voltage": voltage,
    }
    ret.update(
        {
            "max_75V": bool(flags & 0x2),
            "max_100V": bool(flags & 0x4),
            "max_150V": bool(flags & 0x8),
        }
    )
    return ret


@parser(0x0685)
def pz_get_tpz_slewrates(data: bytes) -> Dict[str, Any]:
    chan_ident, slew_open, slew_closed = struct.unpack_from("<HHH", data, HEADER_SIZE)
    return {
        "chan_ident": chan_ident,
        "slew_open": slew_open,
        "slew_closed": slew_closed,
    }


@parser(0x07F2)
def kpz_get_kcubemmiparams(data: bytes) -> Dict[str, Any]:
    (
        _,
        js_mode,
        js_volt_gearbox,
        js_volt_step,
        dir_sense,
        preset_volt1,
        preset_volt2,
        disp_brightness,
        disp_timeout,
        disp_dim_level,
        _,
        _,
        _,
        _,
    ) = struct.unpack_from("<HHHLHLLHHH", data, HEADER_SIZE)
    return {
        "js_mode": js_mode,
        "js_volt_gearbox": js_volt_gearbox,
        "js_volt_step": js_volt_step,
        "dir_sense": dir_sense,
        "preset_volt1": preset_volt1,
        "preset_volt2": preset_volt2,
        "disp_brightness": disp_brightness,
        "disp_timeout": disp_timeout,
        "disp_dim_level": disp_dim_level,
    }


@parser(0x07F5)
def kpz_get_kcubetrigioconfig(data: bytes) -> Dict[str, Any]:
    _, trig1_mode, trig1_polarity, trig2_mode, trig2_polarity, _ = struct.unpack_from(
        "<HHHHHH", data, HEADER_SIZE
    )
    return {
        "trig1_mode": trig1_mode,
        "trig1_polarity": trig1_polarity,
        "trig2_mode": trig2_mode,
        "trig2_polarity": trig2_polarity,
    }


@parser(0x07DC)
def pz_get_tsg_iosettings(data: bytes) -> Dict[str, Any]:
    _, hub_analog_output, display_mode, force_calibration, _, _ = struct.unpack_from(
        "<HHHHHH", data, HEADER_SIZE
    )
    return {
        "hub_analog_output": hub_analog_output,
        "display_mode": display_mode,
        "force_calibration": force_calibration,
    }


@parser(0x07DF)
def pz_get_tsg_reading(data: bytes) -> Dict[str, Any]:
    chan_ident, reading, smoothed = struct.unpack_from("<Hhh", data, HEADER_SIZE)
    return {
        "chan_ident": chan_ident,
        "reading": reading,
        "smoothed": smoothed,
    }


@parser(0x07F8)
def ksg_get_kcubemmiparams(data: bytes) -> Dict[str, Any]:
    chan_ident, disp_intensity, disp_timeout, disp_dim_level = struct.unpack_from(
        "<HHHH", data, HEADER_SIZE
    )
    return {
        "chan_ident": chan_ident,
        "disp_intensity": disp_intensity,
        "disp_timeout": disp_timeout,
        "disp_dim_level": disp_dim_level,
    }


@parser(0x07FB)
def ksg_get_kcubetrigioconfig(data: bytes) -> Dict[str, Any]:
    (
        _,
        trig1_mode,
        trig1_polarity,
        trig2_mode,
        trig2_polarity,
        lower_lim,
        upper_lim,
        smoothing_samples,
        _,
    ) = struct.unpack_from("<HHHHHllHH", data, HEADER_SIZE)
    return {
        "trig1_mode": trig1_mode,
        "trig1_polarity": trig1_polarity,
        "trig2_mode": trig2_mode,
        "trig2_polarity": trig2_polarity,
        "lower_lim": lower_lim,
        "upper_lim": upper_lim,
        "smoothing_samples": smoothing_samples,
    }


@parser(0x0605)
def pz_get_ntmode(data: bytes) -> Dict[str, Any]:
    return {"state": data[2], "mode": data[3]}


@parser(0x0608)
def pz_get_nttrackthreshold(data: bytes) -> Dict[str, Any]:
    threshold_abs_reading = struct.unpack_from("<f", data, HEADER_SIZE)
    return {
        "threshold_abs_reading": threshold_abs_reading,
    }


@parser(0x0611)
def pz_get_ntcirchomepos(data: bytes) -> Dict[str, Any]:
    circ_home_pos_a, circ_home_pos_b = struct.unpack_from("<HH", data, HEADER_SIZE)
    return {
        "circ_home_pos_a": circ_home_pos_a,
        "circ_home_pos_b": circ_home_pos_b,
    }


@parser(0x061A)
def pz_get_ntcircparams(data: bytes) -> Dict[str, Any]:
    (
        circ_dia_mode,
        circ_dia_sw,
        circ_osc_freq,
        abs_pwr_min_circ_dia,
        abs_pwr_max_circ_dia,
        abs_pwr_adjust_type,
    ) = struct.unpack_from("<HHHHHH", data, HEADER_SIZE)
    return {
        "circ_dia_mode": circ_dia_mode,
        "circ_dia_sw": circ_dia_sw,
        "circ_osc_freq": circ_osc_freq,
        "abs_pwr_min_circ_dia": abs_pwr_min_circ_dia,
        "abs_pwr_max_circ_dia": abs_pwr_max_circ_dia,
        "abs_pwr_adjust_type": abs_pwr_adjust_type,
    }


@parser(0x0623)
def pz_get_ntcircdialut(data: bytes) -> Dict[str, Any]:
    lut_val = struct.unpack_from("<16H", data, HEADER_SIZE)
    return {
        "lut_val": lut_val,
    }


@parser(0x0628)
def pz_get_ntphasecompparams(data: bytes) -> Dict[str, Any]:
    phase_comp_mode, phase_comp_asw, phase_comp_bsw = struct.unpack_from(
        "<Hhh", data, HEADER_SIZE
    )
    return {
        "phase_comp_mode": phase_comp_mode,
        "phase_comp_asw": phase_comp_asw,
        "phase_comp_bsw": phase_comp_bsw,
    }


@parser(0x0632)
def pz_get_nttiarangeparams(data: bytes) -> Dict[str, Any]:
    (
        range_mode,
        range_up_limit,
        range_down_limit,
        settle_sample,
        range_change_type,
        range_sw,
    ) = struct.unpack_from("<HhhhHH", data, HEADER_SIZE)
    return {
        "range_mode": range_mode,
        "range_up_limit": range_up_limit,
        "range_down_limit": range_down_limit,
        "settle_sample": settle_sample,
        "range_change_type": range_change_type,
        "range_sw": range_sw,
    }


@parser(0x0635)
def pz_get_ntgainparams(data: bytes) -> Dict[str, Any]:
    gain_ctrl_mode, nt_gain_sw = struct.unpack_from("<Hh", data, HEADER_SIZE)
    return {
        "gain_ctrl_mode": gain_ctrl_mode,
        "nt_gain_sw": nt_gain_sw,
    }


@parser(0x0638)
def pz_get_nttiapfilterparams(data: bytes) -> Dict[str, Any]:
    param_1, _, _, _, _ = struct.unpack_from("<5l", data, HEADER_SIZE)
    return {
        "param_1": param_1,
    }


@parser(0x063A)
def pz_get_nttiareading(data: bytes) -> Dict[str, Any]:
    abs_reading, rel_reading, range, under_over_read = struct.unpack_from(
        "<fHHH", data, HEADER_SIZE
    )
    return {
        "abs_reading": abs_reading,
        "rel_reading": rel_reading,
        "range": range,
        "under_over_read": under_over_read,
    }


@parser(0x063D)
def pz_get_ntfeedbacksrc(data: bytes) -> Dict[str, Any]:
    return {
        "feedback_src": data[2],
    }


@parser(0x063F)
def pz_get_ntstatusbits(data: bytes) -> Dict[str, Any]:
    (status_bits,) = struct.unpack_from("<l", data, HEADER_SIZE)
    return _parse_nt_status_bits(status_bits)


@parser(0x0665)
def pz_get_ntstatusupdate(data: bytes) -> Dict[str, Any]:
    (
        circ_pos_a,
        circ_pos_b,
        circ_dia,
        abs_reading,
        rel_reading,
        range,
        under_over_read,
        status_bits,
        nt_gain,
        phase_comp_a,
        phase_comp_b,
    ) = struct.unpack_from("<HHfHHHlhhh", data, HEADER_SIZE)
    ret = {
        "circ_pos_a": circ_pos_a,
        "circ_pos_b": circ_pos_b,
        "circ_dia": circ_dia,
        "abs_reading": abs_reading,
        "rel_reading": rel_reading,
        "range": range,
        "under_over_read": under_over_read,
        "nt_gain": nt_gain,
        "phase_comp_a": phase_comp_a,
        "phase_comp_b": phase_comp_b,
    }
    ret.update(_parse_nt_status_bits(status_bits))
    return ret


@parser(0x0689)
def kna_get_nttialpfiltercoeefs(data: bytes) -> Dict[str, Any]:
    param_1, _, _, _, _ = struct.unpack_from("<5l", data, HEADER_SIZE)
    return {
        "param_1": param_1,
    }


@parser(0x068C)
def kna_get_kcubemmiparams(data: bytes) -> Dict[str, Any]:
    wheel_step, disp_brightness, _, _, _, _, _, _ = struct.unpack_from(
        "<HH6H", data, HEADER_SIZE
    )
    return {
        "wheel_step": wheel_step,
        "disp_brightness": disp_brightness,
    }


@parser(0x068F)
def kna_get_kcubetrigioconfig(data: bytes) -> Dict[str, Any]:
    t1_mode, t1_polarity, _, t2_mode, t2_polarity, _, _, _, _, _ = struct.unpack_from(
        "<10H", data, HEADER_SIZE
    )
    return {
        "t1_mode": t1_mode,
        "t1_polarity": t1_polarity,
        "t2_mode": t2_mode,
        "t2_polarity": t2_polarity,
    }


@parser(0x06A1)
def kna_get_xyscan(data: bytes) -> Dict[str, Any]:
    line_number, range, *intensity_map = struct.unpack_from("<HH96B", data, HEADER_SIZE)
    return {
        "line_number": line_number,
        "range": range,
        "intensity_map": intensity_map,
    }


@parser(0x07EA)
def nt_get_tna_dispsettings(data: bytes) -> Dict[str, Any]:
    disp_intensity = struct.unpack_from("<H", data, HEADER_SIZE)
    return {
        "disp_intensity": disp_intensity,
    }


@parser(0x07ED)
def nt_get_tnaiosettings(data: bytes) -> Dict[str, Any]:
    lv_out_range, lv_out_route, hv_out_range, sign_io_route = struct.unpack_from(
        "<HHHH", data, HEADER_SIZE
    )
    return {
        "lv_out_range": lv_out_range,
        "lv_out_route": lv_out_route,
        "hv_out_range": hv_out_range,
        "sign_io_route": sign_io_route,
    }


@parser(0x0800)
def la_get_params(data: bytes) -> Dict[str, Any]:
    (submsgid,) = struct.unpack_from("<H", data, HEADER_SIZE)
    ret = {"submsgid": submsgid}
    if submsgid == 1:
        setpoint = struct.unpack_from("<H", data, HEADER_SIZE + 2)
        ret.update({"setpoint": setpoint})
    elif submsgid == 3:
        current, power = struct.unpack_from("<HH", data, HEADER_SIZE + 2)
        ret.update({"current": current, "power": power})
    elif submsgid == 4:
        current, power, voltage = struct.unpack_from("<Hhh", data, HEADER_SIZE + 2)
        ret.update({"current": current, "power": power, "voltage": voltage})
    elif submsgid == 5:
        (laser_source,) = struct.unpack_from("<H", data, HEADER_SIZE + 2)
        ret.update({"laser_source": laser_source})
    elif submsgid == 5:
        (laser_source,) = struct.unpack_from("<H", data, HEADER_SIZE + 2)
        ret.update({"laser_source": laser_source})
    elif submsgid == 7:
        (statusbits,) = struct.unpack_from("<L", data, HEADER_SIZE + 2)
        ret.update(_parse_la_status_bits(statusbits))
    elif submsgid == 9:
        max_current, max_power, wavelength = struct.unpack_from(
            "<HHH", data, HEADER_SIZE + 2
        )
        ret.update(
            {
                "max_current": max_current,
                "max_power": max_power,
                "wavelength": wavelength,
            }
        )
    elif submsgid == 10:
        (max_current,) = struct.unpack_from("<h", data, HEADER_SIZE + 2)
        ret.update({"max_current": max_current})
    elif submsgid == 11:
        intensity, units, _ = struct.unpack_from("<HHH", data, HEADER_SIZE + 2)
        ret.update({"intensity": intensity, "units": units})
    elif submsgid == 13:
        calib_factor, polarity, ramp_up = struct.unpack_from(
            "<fHH", data, HEADER_SIZE + 2
        )
        ret.update(
            {"calib_factor": calib_factor, "polarity": polarity, "ramp_up": ramp_up}
        )
    elif submsgid == 14:
        disp_intensity, _, _, _ = struct.unpack_from("<HHHH", data, HEADER_SIZE + 2)
        ret.update({"disp_intensity": disp_intensity})
    elif submsgid == 17:
        (
            dig_outs,
            _,
        ) = struct.unpack_from("<HH", data, HEADER_SIZE + 2)
        ret.update({"dig_outs": dig_outs})

    return ret


@parser(0x815)
def ld_potrotating(data: bytes):
    return {"degrees": struct.unpack_from("<h", data, 2)}


@parser(0x0819)
def ld_get_maxcurrentdigpot(data: bytes) -> Dict[str, Any]:
    return {
        "max_current": data[2],
    }


@parser(0x0821)
def la_get_statusupdate(data: bytes) -> Dict[str, Any]:
    laser_current, laser_power, statusbits = struct.unpack_from(
        "<HHL", data, HEADER_SIZE
    )
    ret = {
        "laser_current": laser_current,
        "laser_power": laser_power,
    }

    ret.update(_parse_la_status_bits(statusbits))
    return ret


@parser(0x0826)
def ld_get_statusupdate(data: bytes) -> Dict[str, Any]:
    laser_current, photo_current, laser_voltage, _, statusbits = struct.unpack_from(
        "<hHhLL", data, HEADER_SIZE
    )
    ret = {
        "laser_current": laser_current,
        "photo_current": photo_current,
        "laser_voltage": laser_voltage,
    }
    ret.update(_parse_ld_status_bits(statusbits))
    return ret


@parser(0x082C)
def la_get_kcubetrigconfig(data: bytes) -> Dict[str, Any]:
    (
        _,
        trig1_mode,
        trig1_polarity,
        _,
        trig2_mode,
        trig2_polarity,
        _,
    ) = struct.unpack_from("<HHHHHHH", data, HEADER_SIZE)
    return {
        "trig1_mode": trig1_mode,
        "trig1_polarity": trig1_polarity,
        "trig2_mode": trig2_mode,
        "trig2_polarity": trig2_polarity,
    }


@parser(0x0870)
def quad_get_params(data: bytes) -> Dict[str, Any]:
    (submsgid,) = struct.unpack_from("<H", data, HEADER_SIZE)
    ret = {"submsgid": submsgid}
    if submsgid == 1:
        PGain, IGain, DGain = struct.unpack_from("<HHH", data, HEADER_SIZE)
        ret.update(
            {
                "PGain": PGain,
                "IGain": IGain,
                "DGain": DGain,
            }
        )
    elif submsgid == 3:
        x_diff, y_diff, sum, x_pos, y_pos = struct.unpack_from(
            "<hhHhh", data, HEADER_SIZE
        )
        ret.update(
            {
                "x_diff": x_diff,
                "y_diff": y_diff,
                "sum": sum,
                "x_pos": x_pos,
                "y_pos": y_pos,
            }
        )
    elif submsgid == 5:
        (
            x_pos_dem_min,
            y_pos_dem_min,
            x_pos_dem_max,
            y_pos_dem_max,
            lv_out_route,
            ol_pos_dem,
            x_pos_fb_sense,
            y_pos_fb_sense,
        ) = struct.unpack_from("<hhhhHHhh", data, HEADER_SIZE)
        ret.update(
            {
                "x_pos_dem_min": x_pos_dem_min,
                "y_pos_dem_min": y_pos_dem_min,
                "x_pos_dem_max": x_pos_dem_max,
                "y_pos_dem_max": y_pos_dem_max,
                "lv_out_route": lv_out_route,
                "ol_pos_dem": ol_pos_dem,
                "x_pos_fb_sense": x_pos_fb_sense,
                "y_pos_fb_sense": y_pos_fb_sense,
            }
        )
    elif submsgid == 7:
        (mode,) = struct.unpack_from("<H", data, HEADER_SIZE)
        ret.update(
            {
                "mode": mode,
            }
        )
    elif submsgid == 8:
        disp_intensity, disp_mode, disp_dim_timeout = struct.unpack_from(
            "<HHH", data, HEADER_SIZE
        )
        ret.update(
            {
                "disp_intensity": disp_intensity,
                "disp_mode": disp_mode,
                "disp_dim_timeout": disp_dim_timeout,
            }
        )
    elif submsgid == 0xD:
        x_pos, y_pos = struct.unpack_from("<hh", data, HEADER_SIZE)
        ret.update(
            {
                "x_pos": x_pos,
                "y_pos": y_pos,
            }
        )
    elif submsgid == 0xE:
        (
            p,
            i,
            d,
            low_pass_cutoff,
            notch_center,
            filter_q,
            notch_filter_on,
            deriv_filter_on,
        ) = struct.unpack_from("<ffffffHH", data, HEADER_SIZE)
        ret.update(
            {
                "p": p,
                "i": i,
                "d": d,
                "low_pass_cutoff": low_pass_cutoff,
                "notch_center": notch_center,
                "filter_q": filter_q,
                "notch_filter_on": notch_filter_on,
                "deriv_filter_on": deriv_filter_on,
            }
        )
    elif submsgid == 0xF:
        (
            trig1_mode,
            trig1_polarity,
            trig1_sum_min,
            trig1_sum_max,
            trig1_diff_threshold,
            trig2_mode,
            trig2_polarity,
            trig2_sum_min,
            trig2_sum_max,
            trig2_diff_threshold,
            _,
        ) = struct.unpack_from("<11H", data, HEADER_SIZE)
        ret.update(
            {
                "trig1_mode": trig1_mode,
                "trig1_polarity": trig1_polarity,
                "trig1_sum_min": trig1_sum_min,
                "trig1_sum_max": trig1_sum_max,
                "trig1_diff_threshold": trig1_diff_threshold,
                "trig2_mode": trig2_mode,
                "trig2_polarity": trig2_polarity,
                "trig2_sum_min": trig2_sum_min,
                "trig2_sum_max": trig2_sum_max,
                "trig2_diff_threshold": trig2_diff_threshold,
            }
        )
    elif submsgid == 0xA:
        dig_outs, _ = struct.unpack_from("<HH", data, HEADER_SIZE)
        ret.update(
            {
                "dig_outs": dig_outs,
            }
        )

    return ret


@parser(0x0881)
def quad_get_statusupdate(data: bytes) -> Dict[str, Any]:
    x_diff, y_diff, sum, x_pos, y_pos, statusbits = struct.unpack_from(
        "<hhHhhL", data, HEADER_SIZE
    )
    ret = {
        "x_diff": x_diff,
        "y_diff": y_diff,
        "sum": sum,
        "x_pos": x_pos,
        "y_pos": y_pos,
    }
    ret.update(_parse_quad_status_bits(statusbits))
    return ret


@parser(0x0842)
def tec_get_params(data: bytes) -> Dict[str, Any]:
    (submsgid,) = struct.unpack_from("<H", data, HEADER_SIZE)
    ret = {"submsgid": submsgid}
    if submsgid == 1:
        (temp_set,) = struct.unpack_from("<H", data, HEADER_SIZE)
        ret.update({"temp_set": temp_set})
    elif submsgid == 3:
        current, temp_actual, temp_set = struct.unpack_from("<hhH", data, HEADER_SIZE)
        ret.update(
            {
                "current": current,
                "temp_actual": temp_actual,
                "temp_set": temp_set,
            }
        )
    elif submsgid == 5:
        sensor, current_limit = struct.unpack_from("<Hh", data, HEADER_SIZE)
        ret.update(
            {
                "sensor": sensor,
                "current_limit": current_limit,
            }
        )
    elif submsgid == 7:
        (statusbits,) = struct.unpack_from("<L", data, HEADER_SIZE)
        ret.update(_parse_tec_status_bits(statusbits))
    elif submsgid == 9:
        p, i, d = struct.unpack_from("<HHH", data, HEADER_SIZE)
        ret.update(
            {
                "p": p,
                "i": i,
                "d": d,
            }
        )
    elif submsgid == 0xB:
        disp_intensity, disp_mode, _ = struct.unpack_from("<HHH", data, HEADER_SIZE)
        ret.update(
            {
                "disp_intensity": disp_intensity,
                "disp_mode": disp_mode,
            }
        )

    return ret


@parser(0x0861)
def tec_get_statusupdate(data: bytes) -> Dict[str, Any]:
    current, temp_actual, temp_set, statusbits = struct.unpack_from(
        "<hhHL", data, HEADER_SIZE
    )
    ret = {"current": current, "temp_actual": temp_actual, "temp_set": temp_set}
    ret.update(_parse_tec_status_bits(statusbits))
    return ret


@parser(0x08C2)
def pzmot_get_params(data: bytes) -> Dict[str, Any]:
    (submsgid,) = struct.unpack_from("<H", data, HEADER_SIZE)
    ret = {"submsgid": submsgid}
    if submsgid == 5:
        chan_ident, position, _ = struct.unpack_from("<Hll", data, HEADER_SIZE)
        ret.update(
            {
                "chan_ident": chan_ident,
                "position": position,
            }
        )
    elif submsgid == 7:
        chan_ident, max_voltage, step_rate, step_accn = struct.unpack_from(
            "<HHll", data, HEADER_SIZE
        )
        ret.update(
            {
                "chan_ident": chan_ident,
                "max_voltage": max_voltage,
                "step_rate": step_rate,
                "step_accn": step_accn,
            }
        )
    elif submsgid == 9:
        (
            chan_ident,
            jog_mode,
            jog_step_size,
            jog_step_rate,
            jog_step_accn,
        ) = struct.unpack_from("<HHlll", data, HEADER_SIZE)
        ret.update(
            {
                "chan_ident": chan_ident,
                "jog_mode": jog_mode,
                "jog_step_size": jog_step_size,
                "jog_step_rate": jog_step_rate,
                "jog_step_accn": jog_step_accn,
            }
        )
    elif submsgid == 0x11:
        chan_ident, max_step_rate = struct.unpack_from("<Hl", data, HEADER_SIZE)
        ret.update(
            {
                "chan_ident": chan_ident,
                "max_step_rate": max_step_rate,
            }
        )
    elif submsgid == 0x13:
        chan_ident, mode, position1, position2, _, _ = struct.unpack_from(
            "<HHllHH", data, HEADER_SIZE
        )
        ret.update(
            {
                "chan_ident": chan_ident,
                "mode": mode,
                "position1": position1,
                "position2": position2,
            }
        )
    elif submsgid == 0xB:
        chan_ident, fwd_hard_limit, rev_hard_limit, _ = struct.unpack_from(
            "<HHHH", data, HEADER_SIZE
        )
        ret.update(
            {
                "chan_ident": chan_ident,
                "fwd_hard_limit": fwd_hard_limit,
                "rev_hard_limit": rev_hard_limit,
            }
        )
    elif submsgid == 0xF:
        (
            chan_ident,
            home_direction,
            home_lim_switch,
            home_step_rate,
            home_offset_dist,
        ) = struct.unpack_from("<HHHLl", data, HEADER_SIZE)
        ret.update(
            {
                "chan_ident": chan_ident,
                "home_direction": home_direction,
                "home_lim_switch": home_lim_switch,
                "home_step_rate": home_step_rate,
                "home_offset_dist": home_offset_dist,
            }
        )
    elif submsgid == 0x15:
        (
            chan_ident,
            js_mode,
            js_max_step_rate,
            js_dir_sense,
            preset_pos1,
            preset_pos2,
            disp_brightness,
            _,
        ) = struct.unpack_from("<HHlHllHH", data, HEADER_SIZE)
        ret.update(
            {
                "chan_ident": chan_ident,
                "js_mode": js_mode,
                "js_max_step_rate": js_max_step_rate,
                "js_dir_sense": js_dir_sense,
                "preset_pos1": preset_pos1,
                "preset_pos2": preset_pos2,
                "disp_brightness": disp_brightness,
            }
        )
    elif submsgid == 0x17:
        (
            trig_channel1,
            trig_channel2,
            trig1_mode,
            trig1_polarity,
            trig2_mode,
            trig2_polarity,
            *_,
        ) = struct.unpack_from("<12H", data, HEADER_SIZE)
        ret.update(
            {
                "trig_channel1": trig_channel1,
                "trig_channel2": trig_channel2,
                "trig1_mode": trig1_mode,
                "trig1_polarity": trig1_polarity,
                "trig2_mode": trig2_mode,
                "trig2_polarity": trig2_polarity,
            }
        )
    elif submsgid == 0x19:
        (
            chan_ident,
            start_pos_fwd,
            interval_fwd,
            num_pulses_fwd,
            start_pos_reverse,
            interval_rev,
            num_pulses_rev,
            pulse_width,
            num_cycles,
        ) = struct.unpack_from("<Hllllllll", data, HEADER_SIZE)
        ret.update(
            {
                "chan_ident": chan_ident,
                "start_pos_fwd": start_pos_fwd,
                "interval_fwd": interval_fwd,
                "num_pulses_fwd": num_pulses_fwd,
                "start_pos_reverse": start_pos_reverse,
                "interval_rev": interval_rev,
                "num_pulses_rev": num_pulses_rev,
                "pulse_width": pulse_width,
                "num_cycles": num_cycles,
            }
        )
    elif submsgid == 0x2B:
        mode = struct.unpack_from("<H", data, HEADER_SIZE)
        ret.update(
            {
                "mode": mode,
            }
        )
    elif submsgid == 0x2D:
        (
            chan_ident,
            jog_mode,
            jog_step_size_fwd,
            jog_step_size_rev,
            jog_step_rate,
            jog_step_accn,
        ) = struct.unpack_from("<HHllll", data, HEADER_SIZE)
        ret.update(
            {
                "chan_ident": chan_ident,
                "jog_mode": jog_mode,
                "jog_step_size_fwd": jog_step_size_fwd,
                "jog_step_size_rev": jog_step_size_rev,
                "jog_step_rate": jog_step_rate,
                "jog_step_accn": jog_step_accn,
            }
        )
    elif submsgid == 0x30:
        chan_ident, fb_signal_mode, encoder_const = struct.unpack_from(
            "<HHl", data, HEADER_SIZE
        )
        ret.update(
            {
                "chan_ident": chan_ident,
                "fb_signal_mode": fb_signal_mode,
                "encoder_const": encoder_const,
            }
        )
    elif submsgid == 0x32:
        chan_ident, rel_distance = struct.unpack_from("<Hl", data, HEADER_SIZE)
        ret.update(
            {
                "chan_ident": chan_ident,
                "rel_distance": rel_distance,
            }
        )
    elif submsgid == 0x34:
        chan_ident, abs_distance = struct.unpack_from("<Hl", data, HEADER_SIZE)
        ret.update(
            {
                "chan_ident": chan_ident,
                "abs_distance": abs_distance,
            }
        )

    return ret


@parser(0x08D6)
def pzmot_move_completed(data: bytes) -> Dict[str, Any]:
    chan_ident, abs_position, _, _ = struct.unpack_from("<Hlll", data, HEADER_SIZE)
    return {
        "chan_ident": chan_ident,
        "abs_position": abs_position,
    }


@parser(0x08E1)
def pzmot_get_statusupdate(data: bytes) -> Dict[str, Any]:
    chan_ident, position, _, status_bits = struct.unpack_from(
        "<HllL", data, HEADER_SIZE
    )
    ret = {
        "chan_ident": chan_ident,
        "position": position,
    }
    ret.update(_parse_pzmot_status_bits(status_bits))
    return ret


@parser(0x0532)
def pol_get_params(data: bytes) -> Dict[str, Any]:
    _, velocity, home_position, jog_step1, jog_step2, jog_step3 = struct.unpack_from(
        "<HHHHHH", data, HEADER_SIZE
    )
    return {
        "velocity": velocity,
        "home_position": home_position,
        "jog_step1": jog_step1,
        "jog_step2": jog_step2,
        "jog_step3": jog_step3,
    }



# Copyright 2021 Patrick C. Tapping
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ["APTDevice_DC"]

from .. import protocol as apt
from .aptdevice import APTDevice
from ..enums import EndPoint

class APTDevice_DC(APTDevice):
    """
    Initialise and open serial device for a ThorLabs APT controller based on a DC motor drive,
    such as a linear translation stage.

    :param serial_port: Serial port device the device is connected to.
    :param serial_number: Regular expression matching the serial number of device to search for.
    :param home: Perform a homing operation on initialisation.
    :param invert_direction_logic: Invert the meaning of "forward" and "reverse".
    :param swap_limit_switches: Swap "forward" and "reverse" limit switch values.
    :param controller: The destination :class:`EndPoint <thorlabs_apt_device.enums.EndPoint>` for the controller.
    :param bays: Tuple of :class:`EndPoint <thorlabs_apt_device.enums.EndPoint>`\\ (s) for the populated controller bays.
    :param channels: Tuple of indices (1-based) for the controller bay's channels.
    """

    def __init__(self, serial_port=None, serial_number="", home=True, invert_direction_logic=False, swap_limit_switches=False, controller=EndPoint.RACK, bays=(EndPoint.BAY0,), channels=(1,)):

        super().__init__(serial_port=serial_port, serial_number=serial_number, controller=controller, bays=bays, channels=channels)

        self.invert_direction_logic = invert_direction_logic
        """
        On some devices, "forward" velocity moves towards negative encoder counts.
        If that seems opposite to what is expected, this flag allows inversion of that logic.
        This will also swap the meaning of the ``"moving_forward"`` and ``"moving_reverse"`` 
        fields in the :data:`status` flags.
        """

        self.swap_limit_switches = swap_limit_switches
        """
        On some devices, the "forward" limit switch triggers when hitting the limit in the negative
        encoder count direction, and the "reverse" limit switch triggers towards positive encoder 
        counts.
        If that seems opposite to what is expected, this flag swaps the values of the
        ``"forward_limit_switch"`` and ``"reverse_limit_switch"`` fields in the :data:`status`
        flags.

        Note that this is independent of :data:`invert_direction_logic`.
        A stage may report it is "moving forward" (towards either positive or negative encoder
        counts), and then trigger the "reverse" limit switch.
        """

        self.status_ = [[{
            "position" : 0,
            "velocity": 0.0,
            "forward_limit_switch" : False,
            "reverse_limit_switch" : False,
            "moving_forward" : False,
            "moving_reverse" : False,
            "jogging_forward" : False,
            "jogging_reverse" : False,
            "motor_connected" : False,
            "homing" : False,
            "homed" : False,
            "tracking" : False,
            "interlock" : False,
            "settled" : False,
            "motion_error" : False,
            "motor_current_limit_reached" : False,
            "channel_enabled" : False,
            # Update message fields
            "msg" : "",
            "msgid" : 0,
            "source" : 0,
            "dest" : 0,
            "chan_ident" : 0,
        } for _ in self.channels] for _ in self.bays]
        """
        Array of dictionaries of status information for the bays and channels of the device.

        As a device may have multiple card bays, each with multiple channels, this data structure
        is an array of array of dicts. The first axis of the array indexes the bay, the second
        indexes the channel.
        For example, stage.status_[0][1] will return the status dictionary for the first bay, second
        channel.

        Keys for the dictionary are ``"position"``, ``"velocity"``, ``"forward_limit_switch"``,
        ``"reverse_limit_switch"``, ``"moving_forward"``, ``"moving_reverse"``,
        ``"jogging_forward"``, ``"jogging_reverse"``, ``"motor_connected"``, ``"homing"``,
        ``"homed"``, ``"tracking"``, ``"interlock"``, ``"settled"``, ``"motion_error"``,
        ``"motor_current_limit_reached"``, and ``"channel_enabled"``.
        Note that not all keys are relevant to every device.
        
        The documentation states that position is measured in encoder counts, but velocity is
        returned in real units of mm/second.

        Additionally, information about the most recent status message which updated the
        dictionary are also available as ``"msgid"``, ``"source"``, ``"dest"``, and
        ``"chan_ident"``.
        """
        # Status updates are received automatically (configured by APTDevice super class)
        
        self.velparams_ = [[{
            "min_velocity" : 0,
            "max_velocity" : 0,
            "acceleration" : 0,
            # Update message fields
            "msg" : "",
            "msgid" : 0,
            "source" : 0,
            "dest" : 0,
            "chan_ident" : 0,
        } for _ in self.channels] for _ in self.bays]
        """
        Array of dictionaries of velocity parameters.

        As a device may have multiple card bays, each with multiple channels, this data structure
        is an array of array of dicts. The first axis of the array indexes the bay, the second
        indexes the channel.

        Keys are ``"min_velocity"``, ``"max_velocity"``, and ``"acceleration"``.
        Units are encoder counts/second for velocities and counts/second/second for acceleration.
        """
        # Request current velocity parameters
        for bay in self.bays:
            for channel in self.channels:
                self._loop.call_soon_threadsafe(self._write, apt.mot_req_velparams(source=EndPoint.HOST, dest=bay, chan_ident=channel))
        

        self.genmoveparams_ = [[{
            "backlash_distance" : 0,
            # Update message fields
            "msg" : "",
            "msgid" : 0,
            "source" : 0,
            "dest" : 0,
            "chan_ident" : 0,
        } for _ in self.channels] for _ in self.bays]
        """
        Array of dictionaries of general move parameters.

        As a device may have multiple card bays, each with multiple channels, this data structure
        is an array of array of dicts. The first axis of the array indexes the bay, the second
        indexes the channel.

        The only documented parameter is the backlash compensation move distance,
        ``"backlash_distance"``, measured in encoder counts.
        """
        # Request current general move parameters
        for bay in self.bays:
            for channel in self.channels:
                self._loop.call_soon_threadsafe(self._write, apt.mot_req_genmoveparams(source=EndPoint.HOST, dest=bay, chan_ident=channel))


        self.jogparams_ = [[{
            "jog_mode" : 0,
            "step_size" : 0,
            "min_velocity" : 0,
            "acceleration" : 0,
            "max_velocity" : 0,
            "stop_mode" : 0,
            # Update message fields
            "msg" : "",
            "msgid" : 0,
            "source" : 0,
            "dest" : 0,
            "chan_ident" : 0,
        } for _ in self.channels] for _ in self.bays]
        """
        Array of dictionaries of jog parameters.

        As a device may have multiple card bays, each with multiple channels, this data structure
        is an array of array of dicts. The first axis of the array indexes the bay, the second
        indexes the channel.

        Keys are ``"jog_mode"``, ``"step_size"``, ``"min_velocity"``, ``"acceleration"``,
        ``"max_velocity"``, and "stop_mode".
        """
        # Request current jog parameters
        for bay in self.bays:
            for channel in self.channels:
                self._loop.call_soon_threadsafe(self._write, apt.mot_req_jogparams(source=EndPoint.HOST, dest=bay, chan_ident=channel))

        # Home each device if requested
        if home:
            for bay_i, _ in enumerate(self.bays):
                for channel_i, _ in enumerate(self.channels):
                    self.set_enabled(True, bay=bay_i, channel=channel_i)
                    # Sending enabled then home immediately on the TDC001 locks it up.
                    #self.home(bay=bay_i, channel=channel_i)
                    self._loop.call_later(1.0, self.home, bay_i, channel_i)


    def _process_message(self, m):
        super()._process_message(m)
        
        # Decode bay and channel IDs and check if they match one of ours
        if m.msg in ("mot_get_dcstatusupdate", "mot_move_stopped", "mot_move_completed", 
                     "mot_get_velparams",
                     "mot_get_genmoveparams", "mot_genmoveparams",
                     "mot_get_jogparams",
                     "mot_get_avmodes"):
            if m.source == EndPoint.USB:
                # Map USB controller endpoint to first bay
                bay_i = 0
            else:
                # Check if source matches one of our bays
                try:
                    bay_i = self.bays.index(m.source)
                except ValueError:
                    # Ignore message from unknown bay id
                    self._log.warn(f"Message {m.msg} has unrecognised source={m.source}.")
                    return
            # Check if channel matches one of our channels
            try:
                channel_i = self.channels.index(m.chan_ident)
            except ValueError:
                    # Ignore message from unknown channel id
                    self._log.warn(f"Message {m.msg} has unrecognised channel={m.chan_ident}.")
                    return
        
        # Act on each message type
        if m.msg in ("mot_get_dcstatusupdate", "mot_move_stopped", "mot_move_completed"):
            # DC motor status update message    
            self.status_[bay_i][channel_i].update(m._asdict())
            # Scale velocity so it should be in mm/second
            # The explaination of scaling in the documentation doesn't make sense, but
            # dividing the returned value by 2.048 seems sensible (or by 2048 to give m/s)
            self.status_[bay_i][channel_i]["velocity"] /= 2.048
            # Swap meaning of "moving foward" and "moving reverse" if requested
            if self.invert_direction_logic:
                tmp = self.status_[bay_i][channel_i]["moving_forward"]
                self.status_[bay_i][channel_i]["moving_forward"] = self.status_[bay_i][channel_i]["moving_reverse"]
                self.status_[bay_i][channel_i]["moving_reverse"] = tmp
            # Swap values of limit switches if requested
            if self.swap_limit_switches:
                tmp = self.status_[bay_i][channel_i]["forward_limit_switch"]
                self.status_[bay_i][channel_i]["forward_limit_switch"] = self.status_[bay_i][channel_i]["reverse_limit_switch"]
                self.status_[bay_i][channel_i]["reverse_limit_switch"] = tmp
        elif m.msg == "mot_get_velparams":
            # Velocity parameter update
            self.velparams_[bay_i][channel_i].update(m._asdict())
        elif m.msg in ("mot_get_genmoveparams", "mot_genmoveparams"):
            # General move parameter update
            self.genmoveparams_[bay_i][channel_i].update(m._asdict())
        elif m.msg == "mot_get_jogparams":
            # Jog move parameter update
            self.jogparams_[bay_i][channel_i].update(m._asdict())
        else:
            #self._log.debug(f"Received message (unhandled): {m}")
            pass


    def home(self, bay=0, channel=0):
        """
        Home the device.

        :param bay: Index (0-based) of controller bay to send the command.
        :param channel: Index (0-based) of controller bay channel to send the command.
        """
        self._log.debug(f"Homing [bay={self.bays[bay]:#x}, channel={self.channels[channel]}].")
        self._loop.call_soon_threadsafe(self._write, apt.mot_move_home(source=EndPoint.HOST, dest=self.bays[bay], chan_ident=self.channels[channel]))


    def move_relative(self, distance=None, now=True, bay=0, channel=0):
        """
        Perform a relative move.

        :param distance: Movement amount in encoder steps.
        :param now: Perform movement immediately, or wait for subsequent trigger.
        :param bay: Index (0-based) of controller bay to send the command.
        :param channel: Index (0-based) of controller bay channel to send the command.
        """
        if now == True:
            self._log.debug(f"Relative move by {distance} steps [bay={self.bays[bay]:#x}, channel={self.channels[channel]}].")
            self._loop.call_soon_threadsafe(self._write, apt.mot_move_relative(source=EndPoint.HOST, dest=self.bays[bay], chan_ident=self.channels[channel], distance=distance))
        elif now == False and (not distance is None):
            self._log.debug(f"Preparing relative move by {distance} steps [bay={self.bays[bay]:#x}, channel={self.channels[channel]}].")
            self._loop.call_soon_threadsafe(self._write, apt.mot_set_moverelparams(source=EndPoint.HOST, dest=self.bays[bay], chan_ident=self.channels[channel], relative_distance=distance))
        else:
            # Don't move now, and no distance specified...
            self._log.warning("Requested a move_relative with now=False and distance=None: This does nothing!")


    def move_absolute(self, position=None, now=True, bay=0, channel=0):
        """
        Perform an absolute move.

        :param position: Movement destination in encoder steps.
        :param now: Perform movement immediately, or wait for subsequent trigger.
        :param bay: Index (0-based) of controller bay to send the command.
        :param channel: Index (0-based) of controller bay channel to send the command.
        """
        if now == True:
            self._log.debug(f"Absolute move to {position} steps [bay={self.bays[bay]:#x}, channel={self.channels[channel]}].")
            self._loop.call_soon_threadsafe(self._write, apt.mot_move_absolute(source=EndPoint.HOST, dest=self.bays[bay], chan_ident=self.channels[channel], position=position))
        elif now == False and (not position is None):
            self._log.debug(f"Preparing absolute move to {position} steps [bay={self.bays[bay]:#x}, channel={self.channels[channel]}].")
            self._loop.call_soon_threadsafe(self._write, apt.mot_set_moveabsparams(source=EndPoint.HOST, dest=self.bays[bay], chan_ident=self.channels[channel], absolute_position=position))
        else:
            # Don't move now, and no position specified...
            self._log.warning("Requested a move_absolute with now=False and position=None: This does nothing!")


    def stop(self, immediate=False, bay=0, channel=0):
        """
        Stop any current movement.

        :param immediate: Stop immediately, or using the profiled velocity curves.
        :param bay: Index (0-based) of controller bay to send the command.
        :param channel: Index (0-based) of controller bay channel to send the command.
        """
        # False == 2 == profiled, True == 1 == immediate
        stop_mode = (2, 1)[bool(immediate)]
        self._log.debug(f"Stopping {'immediately' if stop_mode == 1 else 'profiled'} [bay={self.bays[bay]:#x}, channel={self.channels[channel]}].")
        self._loop.call_soon_threadsafe(self._write, apt.mot_move_stop(source=EndPoint.HOST, dest=self.bays[bay], chan_ident=self.channels[channel], stop_mode=stop_mode))
    

    def move_velocity(self, direction="forward", bay=0, channel=0):
        """
        Start a movement at constant velocity in the specified direction.

        Direction can be specified as boolean, string or numerical:
        
            * ``False`` is reverse and ``True`` is forward.
            * ``reverse`` is reverse and any other string is forward.
            * ``0`` or ``2`` (or any even number) is reverse and ``1`` (or any odd number) is forward.

        :param direction: Direction to move (forward or reverse).
        :param bay: Index (0-based) of controller bay to send the command.
        :param channel: Index (0-based) of controller bay channel to send the command.
        """
        if type(direction) is bool:
            # False == 2 == reverse, True == 1 == forward
            direction = (2, 1)[direction]
        elif type(direction) is str:
            # forward unless specifically "reverse"
            direction = 2 if direction == "reverse" else 1
        elif type(direction) in (int, float):
            # forward = 1 (or odd numbers), reverse = 0 or 2 (even numbers)
            direction = 2 - int(direction)%2
        else:
            # Otherwise, default to forward
            self._log.warning("Requested a move_velocity with unknown direction \"{direction}\", defaulting to forward.")
            direction = 1
        self._log.debug(f"Velocity move {'forward' if direction == 1 else 'reverse'} [bay={self.bays[bay]:#x}, channel={self.channels[channel]}].")
        # Invert the forward=negative to forward=positive direction logic if requested
        direction = 2 - (direction + bool(self.invert_direction_logic))%2
        self._loop.call_soon_threadsafe(self._write, apt.mot_move_velocity(source=EndPoint.HOST, dest=self.bays[bay], chan_ident=self.channels[channel], direction=direction))


    def set_velocity_params(self, acceleration, max_velocity, bay=0, channel=0):
        """
        Configure the parameters for movement velocity.

        :param acceleration: Acceleration in counts/second/second.
        :param max_velocity: Maximum velocity in counts/second.
        :param bay: Index (0-based) of controller bay to send the command.
        :param channel: Index (0-based) of controller bay channel to send the command.
        """
        self._log.debug(f"Setting velocity parameters to accel={acceleration}, max={max_velocity} [bay={self.bays[bay]:#x}, channel={self.channels[channel]}].")
        self._loop.call_soon_threadsafe(self._write, apt.mot_set_velparams(source=EndPoint.HOST, dest=self.bays[bay], chan_ident=self.channels[channel], min_velocity=0, acceleration=acceleration, max_velocity=max_velocity))
        # Update status with new velocity parameters
        self._loop.call_soon_threadsafe(self._write, apt.mot_req_velparams(source=EndPoint.HOST, dest=self.bays[bay], chan_ident=self.channels[channel]))


    def set_enabled(self, state=True, bay=0, channel=0):
        """
        Enable or disable a device.

        :param state: Set to ``True`` for enable, ``False`` for disabled.
        :param bay: Index (0-based) of controller bay to send the command.
        :param channel: Index (0-based) of controller bay channel to send the command.
        """
        state = (2, 1)[bool(state)]
        self._log.debug(f"Setting channel enabled={state == 1} [bay={self.bays[bay]:#x}, channel={self.channels[channel]}].")
        self._loop.call_soon_threadsafe(self._write, apt.mod_set_chanenablestate(source=EndPoint.HOST, dest=self.bays[bay], chan_ident=self.channels[channel], enable_state=state))


    def set_jog_params(self, size, acceleration, max_velocity, continuous=False, immediate_stop=False, bay=0, channel=0):
        """
        Configure the parameters for jogging movements.

        :param size: Size of movement in encoder counts.
        :param acceleration: Acceleration in counts/second/second.
        :param max_velocity: Maximum velocity in counts/second.
        :param continuous: Continuous movement, or single step.
        :param immediate_stop: Stop immediately, or using the profiled velocity curves.
        :param bay: Index (0-based) of controller bay to send the command.
        :param channel: Index (0-based) of controller bay channel to send the command.
        """
        # False == 2 == profiled, True == 1 == immediate
        stop_mode = (2, 1)[bool(immediate_stop)]
        # False == 2 == stepped, True == 1 == continuous
        jog_mode = (2, 1)[bool(continuous)]
        self._log.debug(f"Setting jog parameters to size={size}, accel={acceleration}, max={max_velocity}, cont={continuous}, imm={immediate_stop} [bay={self.bays[bay]:#x}, channel={self.channels[channel]}].")
        self._loop.call_soon_threadsafe(self._write, apt.mot_set_jogparams(source=EndPoint.HOST, dest=self.bays[bay], chan_ident=self.channels[channel], step_size=size, min_velocity=0, acceleration=acceleration, max_velocity=max_velocity, jog_mode=jog_mode, stop_mode=stop_mode))
        # Update status with new jog parameters
        self._loop.call_soon_threadsafe(self._write, apt.mot_req_jogparams(source=EndPoint.HOST, dest=self.bays[bay], chan_ident=self.channels[channel]))


    def move_jog(self, direction="forward", bay=0, channel=0):
        """
        Start a jog movement in the specified direction.

        Direction can be specified as boolean, string or numerical:
        
            * ``False`` is reverse and ``True`` is forward.
            * ``reverse`` is reverse and any other string is forward.
            * ``0`` or ``2`` (or any even number) is reverse and ``1`` (or any odd number) is forward.

        :param direction: Direction to move (forward or reverse).
        :param bay: Index (0-based) of controller bay to send the command.
        :param channel: Index (0-based) of controller bay channel to send the command.
        """
        if type(direction) is bool:
            # False == 2 == reverse, True == 1 == forward
            direction = (2, 1)[direction]
        elif type(direction) is str:
            # forward unless specifically "reverse"
            direction = 2 if direction == "reverse" else 1
        elif type(direction) in (int, float):
            # forward = 1 (or odd numbers), reverse = 0 or 2 (even numbers)
            direction = 2 - int(direction)%2
        else:
            # Otherwise, default to forward
            self._log.warning("Requested a move_jog with unknown direction \"{direction}\", defaulting to forward.")
            direction = 1
        self._log.debug(f"Jog move {'forward' if direction == 1 else 'reverse'} [bay={self.bays[bay]:#x}, channel={self.channels[channel]}].")
        # Invert the forward=negative to forward=positive direction logic if requested
        direction = 2 - (direction + bool(self.invert_direction_logic))%2
        self._loop.call_soon_threadsafe(self._write, apt.mot_move_jog(source=EndPoint.HOST, dest=self.bays[bay], chan_ident=self.channels[channel], direction=direction))


    def set_move_params(self, backlash_distance, bay=0, channel=0):
        """
        Set parameters for generic move commands, currently only the backlash compensation distance.

        :param backlash_distance: Backlash compensation distance in encoder counts.
        :param bay: Index (0-based) of controller bay to send the command.
        :param channel: Index (0-based) of controller bay channel to send the command.
        """
        self._log.debug(f"Setting move parameters to backlash={backlash_distance} steps [bay={self.bays[bay]:#x}, channel={self.channels[channel]}].")
        self._loop.call_soon_threadsafe(self._write, apt.mot_set_genmoveparams(source=EndPoint.HOST, dest=self.bays[bay], chan_ident=self.channels[channel], backlash_distance=backlash_distance))
        # Update status with new general move parameters
        self._loop.call_soon_threadsafe(self._write, apt.mot_req_genmoveparams(source=EndPoint.HOST, dest=self.bays[bay], chan_ident=self.channels[channel]))



# Copyright 2021 Patrick C. Tapping
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ["BBD"]

from .. import protocol as apt
from .aptdevice_dc import APTDevice_DC
from ..enums import EndPoint

class BBD(APTDevice_DC):
    """
    A class for ThorLabs APT device models BBD10x and BBD20x, where x is the number of channels (1, 2 or 3).

    It is based off :class:`APTDevice_DC` with some customisation for the specifics of the device.

    Note that the BBDs are referred to as a x-channel controller, but the actual device layout is that 
    the controller is a "rack" system with three bays, where x number of single-channel
    controller cards may be installed. In other words, the BBD203 "3 channel" controller actually
    has 3 populated bays (``bays=(EndPoint.BAY0, EndPoint.BAY1, EndPoint.BAY2)``), each of which
    only controls a single channel (``channels=(1,)``).

    The parameter ``x`` configures the number of channels.
    If ``x=1`` it is a single bay/channel controller, and aliases of ``status = status_[0][0]``
    etc are created for convenience.

    :param x: Number of channels the device controls.
    :param serial_port: Serial port device the device is connected to.
    :param serial_number: Regular expression matching the serial number of device to search for.
    :param home: Perform a homing operation on initialisation.
    :param invert_direction_logic: Invert the meaning of "forward" and "reverse" directions.
    :param swap_limit_switches: Swap "forward" and "reverse" limit switch values.
    """
    def __init__(self, serial_port=None, serial_number="73", x=1, home=True, invert_direction_logic=False, swap_limit_switches=True):
        
        # Configure number of bays
        if x == 3:
            bays = (EndPoint.BAY0, EndPoint.BAY1, EndPoint.BAY2)
        elif x == 2:
            bays = (EndPoint.BAY0, EndPoint.BAY1)
        else:
            bays = (EndPoint.BAY0,)

        super().__init__(serial_port=serial_port, serial_number=serial_number, home=home, invert_direction_logic=invert_direction_logic, swap_limit_switches=swap_limit_switches, controller=EndPoint.RACK, bays=bays, channels=(1,))
        
        self.trigger_ = [[{
            # Actual integer code returned by device
            "mode" : 0,
            # Unpacked meaning of mode bits
            "input_edge" : "",
            "input_mode" : "",
            "output_logic" : "",
            "output_mode" : "",
            # Update message fields
            "msg" : "",
            "msgid" : 0,
            "source" : 0,
            "dest" : 0,
            "chan_ident" : 0,
        } for _ in self.channels] for _ in self.bays]
        """
        Input and output triggering configuration.

        Fields are ``"input_edge"``, ``"input_mode"``, ``"output_logic"``, and ``"output_mode"``.
        Additionally, ``"mode"`` stores the raw bit field data from the device as an integer.
        """
        # Request current trigger modes
        for bay in self.bays:
            for channel in self.channels:
                self._loop.call_soon_threadsafe(self._write, apt.mot_req_trigger(source=EndPoint.HOST, dest=bay, chan_ident=channel))

        if x == 1:
            self.status = self.status_[0][0]
            """Alias to first bay/channel of :data:`APTDevice_DC.status_`."""
            
            self.velparams = self.velparams_[0][0]
            """Alias to first bay/channel of :data:`APTDevice_DC.velparams_`"""
            
            self.genmoveparams = self.genmoveparams_[0][0]
            """Alias to first bay/channel of :data:`APTDevice_DC.genmoveparams_`"""
            
            self.jogparams = self.jogparams_[0][0]
            """Alias to first bay/channel of :data:`APTDevice_DC.jogparams_`"""
            
            self.trigger = self.trigger_[0][0]
            """Alias to first bay/channel of :data:`BBD.trigger_`"""


    def _process_message(self, m):
        super()._process_message(m)
       
        # Decode bay and channel IDs and check if they match one of ours
        if m.msg in ("mot_get_trigger",):
            # Check if source matches one of our bays
            try:
                bay_i = self.bays.index(m.source)
            except ValueError:
                # Ignore message from unknown bay id
                self._log.warn(f"Message {m.msg} has unrecognised source={m.source}.")
                return
            # Check if channel matches one of our channels
            try:
                channel_i = self.channels.index(m.chan_ident)
            except ValueError:
                    # Ignore message from unknown channel id
                    self._log.warn(f"Message {m.msg} has unrecognised channel={m.chan_ident}.")
                    return
        
        if m.msg == "mot_get_trigger":
            # Trigger mode update message
            self.trigger_[bay_i][channel_i].update(m._asdict())
            # Decode the input bit fields
            self.trigger_[bay_i][channel_i]["input_edge"] = ("falling", "rising")[bool(m.mode & 0b00000001)]
            if m.mode & 0b00000010:
                self.trigger_[bay_i][channel_i]["input_mode"] = "move_relative"
            elif m.mode & 0b00000100:
                self.trigger_[bay_i][channel_i]["input_mode"] = "move_absolute"
            elif m.mode & 0b00001000:
                self.trigger_[bay_i][channel_i]["input_mode"] = "home"
            else:
                self.trigger_[bay_i][channel_i]["input_mode"] = "disabled"
            # Decode the output bit fields
            self.trigger_[bay_i][channel_i]["output_logic"] = ("low", "high")[bool(m.mode & 0b00010000)]
            if m.mode & 0b00100000:
                self.trigger_[bay_i][channel_i]["output_mode"] = "in_motion"
            elif m.mode & 0b01000000:
                self.trigger_[bay_i][channel_i]["output_mode"] = "motion_complete"
            elif m.mode & 0b10000000:
                self.trigger_[bay_i][channel_i]["output_mode"] = "max_velocity"
            else:
                self.trigger_[bay_i][channel_i]["output_mode"] = "disabled"


    def set_trigger(self, input_edge=None, input_mode=None, output_logic=None, output_mode=None, bay=0, channel=0):
        """
        Configure the external input and/or output triggering modes for the device.
        
        The input triggering logic is selected with ``input_edge``:
        
            * ``"rising"`` : Trigger on rising edge (low to high transition).
            * ``"falling"`` : Trigger on falling edge (high to low transition).
    
        On detection of the input trigger, either a relative, absolute, or homing move can be 
        started.
        For the moves, the movement parameters should have been pre-configured using the relevant
        :meth:`move_relative()` or :meth:`move_absolute()` methods, passing the ``now=False``
        parameter to indicate the move should be delayed until a subsequent trigger.
        The movement type once triggered is selected with ``input_mode``:

            * ``"move_relative"`` : Perform a relative move.
            * ``"move_absolute"`` : Perform an absolute move.
            * ``"home"`` : Perform a homing operation.
            * ``"disabled"`` : Do nothing.
        
        The output triggering logic is selected with ``output_logic``:

            * ``"high"`` : Output is high during the event.
            * ``"low"`` : Output is low during the event.

        The event for which the output trigger signal is sent is selected by ``output_mode``:

            * ``"in_motion"`` : Trigger activated during motion.
            * ``"motion_complete"`` : Trigger activated when motion is completed.
            * ``"max_velocity"`` : Trigger activated when movement reaches full speed.
            * ``"disabled"`` : Don't activate trigger.
        
        :param input_edge: Signal edge for input triggering.
        :param input_mode: Action to take on input trigger.
        :param output_logic: Output trigger logic level.
        :param output_mode: Event for which output trigger is activated.
        """
        # Note that all of the "mode" field is treated as a bit mask, which implies/allows for
        # multiple triggering actions. I'm assuming that is not actually possible, so the way
        # this is coded allows for only a single action on trigger.
        
        # TODO, actually use the current mode once that's implemented
        mode = 0

        # Set or clear the TRIGIN_HIGH bit
        # The documentation for this bit is horrible! We'll guess what it's supposed to mean...
        if input_edge == "rising":
            mode |= 0b00000001
        elif input_edge == "falling":
            mode &= 0b11111110
        elif input_edge is None:
            # Leave as is
            pass
        else:
            self._log.warn(f"Unrecognised trigger input_edge '{input_edge}'', should be 'rising' or 'falling'.")
        
        # Set one (only) of the TRIGIN_RELMOVE, TRIGIN_ABSMOVE or TRIGIN_HOMEMOVE bits
        if input_mode == "move_relative":
            mode &= 0b11110001
            mode |= 0b00000010
        elif input_mode == "move_absolute":
            mode &= 0b11110001
            mode |= 0b00000100
        elif input_mode == "home":
            mode &= 0b11110001
            mode |= 0b00001000
        elif input_mode == "disabled":
            mode &= 0b11110001
        elif input_mode is None:
            # Leave as is
            pass
        else:
            self._log.warn(f"Unrecognised trigger input_mode '{input_mode}', should be 'move_relative', 'move_absolute' or 'home'.")
        
        # Set or clear the TIGOUT_HIGH bit
        if output_logic == "high":
            mode |= 0b00010000
        elif output_logic == "low":
            mode &= 0b11101111
        elif output_logic is None:
            # Leave as is
            pass
        else:
            self._log.warn(f"Unrecognised trigger output_logic '{output_logic}', should be 'high' or 'low'")
        
        # Set one (only) of TRIGOUT_INMOTION, TRIGOUT_MOTIONCOMPLETE or TRIGOUT_MAXVELOCITY bits
        if output_mode == "in_motion":
            mode &= 0b00011111
            mode |= 0b00100000
        elif output_mode == "motion_complete":
            mode &= 0b00011111
            mode |= 0b01000000
        elif output_mode == "max_velocity":
            mode &= 0b00011111
            mode |= 0b10000000
        elif output_mode == "disabled":
            mode &= 0b00011111
        elif output_mode is None:
            # Leave as is
            pass
        else:
            self._log.warn(f"Unrecognised trigger output_mode '{output_mode}', should be 'in_motion', 'motion_complete' or 'max_velocity'.")
        
        self._log.debug(f"Setting trigger mode={mode} [bay={self.bays[bay]:#x}, channel={self.channels[channel]}].")
        self._loop.call_soon_threadsafe(self._write, apt.mot_set_trigger(source=EndPoint.HOST, dest=self.bays[bay], chan_ident=self.channels[channel], mode=mode))
        # Update status with new trigger parameters
        self._loop.call_soon_threadsafe(self._write, apt.mot_req_trigger(source=EndPoint.HOST, dest=self.bays[bay], chan_ident=self.channels[channel]))



